# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals

"""
This module contains objects that are used to describe the environments in a structure. The most detailed object
(StructureEnvironments) contains a very thorough analysis of the environments of a given atom but is difficult to
used as such. The LightStructureEnvironments object is a lighter version that is obtained by applying a "strategy"
on the StructureEnvironments object. Basically, the LightStructureEnvironments provides the coordination environment(s)
and possibly some fraction corresponding to these.
"""

__author__ = "David Waroquiers"
__copyright__ = "Copyright 2012, The Materials Project"
__credits__ = "Geoffroy Hautier"
__version__ = "2.0"
__maintainer__ = "David Waroquiers"
__email__ = "david.waroquiers@gmail.com"
__date__ = "Feb 20, 2016"


import numpy as np
from pymatgen.core.sites import PeriodicSite
from monty.json import MSONable, MontyDecoder
from pymatgen.core.periodic_table import Element, Specie
from pymatgen.core.structure import Structure
from monty.json import jsanitize
from pymatgen.analysis.chemenv.coordination_environments.voronoi import DetailedVoronoiContainer
from pymatgen.analysis.chemenv.utils.chemenv_errors import ChemenvError
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometries import AllCoordinationGeometries
from pymatgen.analysis.chemenv.utils.defs_utils import AdditionalConditions

allcg = AllCoordinationGeometries()
symbol_cn_mapping = allcg.get_symbol_cn_mapping()

class StructureEnvironments(MSONable):
    """
    Class used to store the chemical environments of a given structure.
    """
    AC = AdditionalConditions()

    def __init__(self, voronoi, bva_valences, sites_map, equivalent_sites,
                 ce_list, structure):
        """
        Constructor for the StructureEnvironments object.
        :param voronoi: VoronoiContainer object for the structure
        :param bva_valences: Valences obtained using the bond-valence analysis
        :param sites_map: Mapping of equivalent sites to the unequivalent sites that have been computed.
        :param equivalent_sites: List of list of equivalent sites of the structure
        :param struct_sites_to_irreducible_site_list_map: Maps the index of a site to the index of the item in the
        list of equivalent sites to which the site belongs.
        :param ce_list: List of chemical environments
        :param structure: Structure object
        """
        self.voronoi = voronoi
        self.bva_valences = bva_valences
        self.sites_map = sites_map
        self.equivalent_sites = equivalent_sites
        #self.struct_sites_to_irreducible_site_list_map = struct_sites_to_irreducible_site_list_map
        self.ce_list = ce_list
        self.structure = structure

    def get_csm(self, isite, mp_symbol):
        csms = self.get_csms(isite, mp_symbol)
        if len(csms) != 1:
            raise ChemenvError('StructureEnvironments',
                               'get_csm',
                               'Number of csms for site #{} with '
                               'mp_symbol "{}" = {}'.format(str(isite),
                                                            mp_symbol,
                                                            str(len(csms))))
        return csms[0]

    def get_csms(self, isite, mp_symbol):
        """
        Returns the continuous symmetry measure(s) of site with index isite with respect to the
         perfect coordination environment with mp_symbol. For some environments, a given mp_symbol might not
         be available (if there is no voronoi parameters leading to a number of neighbours corresponding to
         the coordination number of environment mp_symbol). For some environments, a given mp_symbol might
         lead to more than one csm (when two or more different voronoi parameters lead to different neighbours
         but with same number of neighbours).
        :param isite: Index of the site
        :param mp_symbol: MP symbol of the perfect environment for which the csm has to be given
        :return: List of csms for site isite with respect to geometry mp_symbol
        """
        cn = symbol_cn_mapping[mp_symbol]
        if cn not in self.ce_list[isite]:
            return []
        else:
            return [envs[mp_symbol] for envs in self.ce_list[isite][cn]]

    def get_environments_figure(self, isite, plot_type=None, title='Coordination numbers', max_dist=2.0,
                                additional_condition=AC.ONLY_ACB, colormap=None, figsize=None):
        """
        Plotting of the coordination environments of a given site for all the distfactor/angfactor regions. The
        chemical environments with the lowest continuous symmetry measure is shown for each distfactor/angfactor
        region as the value for the color of that distfactor/angfactor region (using a colormap).
        :param isite: Index of the site for which the plot has to be done
        :param plot_type: How to plot the coordinations
        :param title: Title for the figure
        :param max_dist: Maximum distance to be plotted when the plotting of the distance is set to 'initial_normalized'
                         or 'initial_real' (Warning: this is not the same meaning in both cases! In the first case,
                         the closest atom lies at a "normalized" distance of 1.0 so that 2.0 means refers to this
                         normalized distance while in the second case, the real distance is used)
        :param figsize: Size of the figure to be plotted
        :return: The figure object to be plotted or saved to file
        """
        try:
            import matplotlib.pyplot as mpl
            from matplotlib import cm
            from matplotlib.colors import Normalize, LinearSegmentedColormap, ListedColormap
            from matplotlib.patches import Rectangle, Polygon
        except ImportError:
            print('Plotting Chemical Environments requires matplotlib ... exiting "plot" function')
            return

        #Initializes the figure
        if figsize is None:
            fig = mpl.figure()
        else:
            fig = mpl.figure(figsize=figsize)
        subplot = fig.add_subplot(111)

        #Initializes the distance and angle parameters
        if plot_type is None:
            plot_type = {'distance_parameter': ('initial_inverse_opposite', None),
                         'angle_parameter': ('initial_opposite', None)}
        bounds_and_limits = self.voronoi.voronoi_parameters_bounds_and_limits(isite, plot_type, max_dist)
        # distance_bounds = bounds_and_limits['distance_bounds']
        # angle_bounds = bounds_and_limits['angle_bounds']
        # dist_limits = bounds_and_limits['distance_limits']
        # ang_limits = bounds_and_limits['angle_limits']
        # iac = self.voronoi.additional_conditions.index(additional_condition)
        if colormap is None:
            mycm = cm.jet
        else:
            mycm = colormap
        mymin = 0.0
        mymax = 10.0
        norm = Normalize(vmin=mymin, vmax=mymax)
        scalarmap = cm.ScalarMappable(norm=norm, cmap=mycm)

        #Plot the rectangles and coordinations
        maps_and_vertices = self.voronoi.maps_and_surface_vertices(isite,
                                                                   additional_condition=additional_condition,
                                                                   plot_type=plot_type, max_dist=max_dist)
        if maps_and_vertices is None:
            return None
        cn_maps = maps_and_vertices['cn_maps']
        bounds_and_limits = maps_and_vertices['bounds_and_limits']
        dist_limits = bounds_and_limits['distance_limits']
        ang_limits = bounds_and_limits['angle_limits']
        vertices_dist_ang = maps_and_vertices['vertices_dist_ang']
        text_info_dist_ang = maps_and_vertices['text_info_dist_ang']
        for i_cn_map, cn_map in enumerate(cn_maps):
            ce = self.ce_list[isite][cn_map[0]][cn_map[1]]
            mingeom = ce.minimum_geometry()
            if mingeom is not None:
                mp_symbol = mingeom[0]
                csm = mingeom[1]['symmetry_measure']
                mycolor = scalarmap.to_rgba(csm)
                myinvcolor = [1.0 - mycolor[0], 1.0 - mycolor[1], 1.0 - mycolor[2], 1.0]
                mytext = '{}'.format(mp_symbol)
            else:
                cn = cn_map[0]
                mycolor = 'w'
                myinvcolor = 'k'
                mytext = '{:d}'.format(cn)
            polygon = Polygon(vertices_dist_ang[i_cn_map], closed=True, edgecolor='k', facecolor=mycolor)
            subplot.add_patch(polygon)
            subplot.annotate(mytext,
                             xy=text_info_dist_ang[i_cn_map],
                             ha='center', va='center', color=myinvcolor, fontsize='x-small')

        title += '\nDist: {}, Ang: {}'.format(plot_type['distance_parameter'][0], plot_type['angle_parameter'][0])
        subplot.set_title(title)
        subplot.set_xlabel('Distance factor')
        subplot.set_ylabel('Angle factor')
        subplot.set_xlim(dist_limits)
        subplot.set_ylim(ang_limits)
        # ax2 = subplot.twin()  # ax2 is responsible for "top" axis and "right" axis
        # ax2.set_xticks([0., .5*np.pi, np.pi, 1.5*np.pi, 2*np.pi])
        # ax2.set_xticklabels(["$0$", r"$\frac{1}{2}\pi$",
        #                      r"$\pi$", r"$\frac{3}{2}\pi$", r"$2\pi$"])
        #
        # ax2.axis["right"].major_ticklabels.set_visible(False)
        scalarmap.set_array([mymin, mymax])
        cb = fig.colorbar(scalarmap, ax=subplot, extend='max')
        cb.set_label('Continuous symmetry measure')
        return fig

    def plot_environments(self,  isite, plot_type=None, title='Coordination numbers', max_dist=2.0,
                          additional_condition=AC.ONLY_ACB, figsize=None):
        """
        Plotting of the coordination numbers of a given site for all the distfactor/angfactor parameters. If the
        chemical environments are given, a color map is added to the plot, with the lowest continuous symmetry measure
        as the value for the color of that distfactor/angfactor set.
        :param isite: Index of the site for which the plot has to be done
        :param plot_type: How to plot the coordinations
        :param title: Title for the figure
        :param max_dist: Maximum distance to be plotted when the plotting of the distance is set to 'initial_normalized'
                         or 'initial_real' (Warning: this is not the same meaning in both cases! In the first case,
                         the closest atom lies at a "normalized" distance of 1.0 so that 2.0 means refers to this
                         normalized distance while in the second case, the real distance is used)
        :param figsize: Size of the figure to be plotted
        :return: Nothing returned, just plot the figure
        """
        fig = self.get_environments_figure(isite=isite, plot_type=plot_type, title=title, max_dist=max_dist,
                                           additional_condition=additional_condition, figsize=figsize)
        if fig is None:
            return
        fig.show()

    def save_environments_figure(self,  isite, imagename='image.png', plot_type=None, title='Coordination numbers',
                                 max_dist=2.0, additional_condition=AC.ONLY_ACB, figsize=None):
        fig = self.get_environments_figure(isite=isite, plot_type=plot_type, title=title, max_dist=max_dist,
                                           additional_condition=additional_condition, figsize=figsize)
        if fig is None:
            return
        fig.savefig(imagename)

    def to_bv_dict(self, isite):
        out = {"neighbours_lists": [], "continuous_symmetry_measures": []}
        for cn, coordnbs_list in self.voronoi._unique_coordinated_neighbors[isite].items():
            for i_coordnbs, coordnbs in enumerate(coordnbs_list):
                neighbours = []
                for ineighb, neighb in enumerate(coordnbs[0]):
                    mydict = {'neighbour': neighb.as_dict(),
                              'distance': coordnbs[2][ineighb]['distance'],
                              'normalized_distance': coordnbs[2][ineighb]['weighted_distance'],
                              'angle': coordnbs[2][ineighb]['angle'],
                              'normalized_angle': coordnbs[2][ineighb]['weighted_angle']}
                    neighbours.append(mydict)
                ce = self.ce_list[isite][cn][i_coordnbs]
                mingeoms = ce.minimum_geometries()
                csm_dict = {ce: ce_dict['symmetry_measure'] for ce, ce_dict in mingeoms}
                out["neighbours_lists"].append(neighbours)
                out["continuous_symmetry_measures"].append(csm_dict)
        return out
      #           - "neighbour": the PeriodicSite object, as a dictionary
      # - "distance": the distance to the site X
      # - "angle": the angle from the site X
      # - "normalized_distance": the normalized distance to the site X
      # - "normalized_angle": the normalized angle from the site X

    def unique_coordinated_neighbors(self, isite=None, cn_map=None):
        return self.voronoi.unique_coordinated_neighbors(isite=isite, cn_map=cn_map)

    def __eq__(self, other):
        if len(self.ce_list) != len(other.ce_list):
            return False
        for ii in range(len(self.ce_list)):
            if self.ce_list[ii] is not None:
                if other.ce_list[ii] is None:
                    return False
                if set(self.ce_list[ii].keys()) != set(other.ce_list[ii].keys()):
                    return False
                for cn, ces in list(self.ce_list[ii].items()):
                    if ces != other.ce_list[ii][cn]:
                        return False
            else:
                if other.ce_list[ii] is not None:
                    return False
        return (self.voronoi == other.voronoi and self.bva_valences == other.bva_valences and
                self.sites_map == other.sites_map and self.equivalent_sites == other.equivalent_sites and
                self.ce_list == other.ce_list and self.structure == other.structure)

    def as_dict(self):
        """
        Bson-serializable dict representation of the StructureEnvironments object.
        :return: Bson-serializable dict representation of the StructureEnvironments object.
        """
        ce_list_dict = [{str(cn): [ce.as_dict() for ce in ce_dict[cn]]
                         for cn in ce_dict} if ce_dict is not None else None for ce_dict in self.ce_list]
        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "voronoi": self.voronoi.as_dict(),
                "bva_valences": self.bva_valences,
                "sites_map": self.sites_map,
                "equivalent_sites": [[ps.as_dict() for ps in psl] for psl in self.equivalent_sites],
                "ce_list": ce_list_dict,
                "structure": self.structure.as_dict()}

    @classmethod
    def from_dict(cls, d):
        """
        Reconstructs the StructureEnvironments object from a dict representation of the StructureEnvironments created
        using the as_dict method.
        :param d: dict representation of the StructureEnvironments object
        :return: StructureEnvironments object
        """
        ce_list = [None if (ce_dict == 'None' or ce_dict is None) else {
            int(cn): [ChemicalEnvironments.from_dict(ced) for ced in ce_dict[cn]]
            for cn in ce_dict} for ce_dict in d['ce_list']]
        return cls(DetailedVoronoiContainer.from_dict(d['voronoi']), d['bva_valences'], d['sites_map'],
                   [[PeriodicSite.from_dict(psd) for psd in psl] for psl in d['equivalent_sites']],
                   ce_list, Structure.from_dict(d['structure']))


class LightStructureEnvironments(MSONable):
    """
    Class used to store the chemical environments of a given structure obtained from a given ChemenvStrategy. Currently,
    only strategies leading to the determination of a unique environment for each site is allowed
    This class does not store all the information contained in the StructureEnvironments object, only the coordination
    environment found
    """
    DELTA_MAX_OXIDATION_STATE = 0.1
    DEFAULT_STATISTICS_FIELDS = ['anion_list', 'anion_atom_list', 'cation_list', 'cation_atom_list',
                                 'neutral_list', 'neutral_atom_list',
                                 'atom_coordination_environments_present',
                                 'bva_ion_coordination_environments_present',
                                 'structure_ion_coordination_environments_present',
                                 'ion_coordination_environments_present',
                                 'fraction_atom_coordination_environments_present',
                                 'fraction_bva_ion_coordination_environments_present',
                                 'fraction_structure_ion_coordination_environments_present',
                                 'fraction_ion_coordination_environments_present',
                                 'coordination_environments_atom_present',
                                 'coordination_environments_ion_present']

    def __init__(self, strategy, structure_environments=None, bva_valences=None,
                 coordination_environments=None, neighbors=None, structure=None, setup_neighbors_by_indices=False,
                 neighbors_by_indices=None):
        """
        Constructor for the LightStructureEnvironments object.
        """
        self._strategy = strategy
        self._uniquely_determined_coordination_environments = strategy.uniquely_determines_coordination_environments
        self.statistics_dict = None
        self._neighbors_by_indices = None

        if coordination_environments is not None:
            self._coordination_environments = coordination_environments
            self._neighbors = neighbors
            self._bva_valences = bva_valences
            self._structure = structure
            if neighbors_by_indices is not None:
                self._neighbors_by_indices = neighbors_by_indices
            if setup_neighbors_by_indices:
                self._setup_neighbors_by_indices()
            return
        self._coordination_environments = []
        self._neighbors = []

        if structure_environments is None:
            raise RuntimeError("coordination_environments and structure_environments are both None")
        self._strategy.set_structure_environments(structure_environments)
        self._structure = structure_environments.structure
        self._bva_valences = structure_environments.bva_valences if bva_valences is None else bva_valences
        for isite, site in enumerate(structure_environments.structure):
            if structure_environments.ce_list[isite] is None:
                self._coordination_environments.append([])
                self._neighbors.append({})
                continue
            site_ce_list = self._strategy.get_site_ce_fractions_and_neighbors(site)
            if site_ce_list is None:
                self._coordination_environments.append([])
                self._neighbors.append({})
                continue
            #This contains a dictionary with keys "ce" and "neighbors".
            #  In site_ce_list["ce"], a list of dictionaries with each possible environment determined by the strategy
            #    In each dictionary, the following keys exist : mp_symbol, fraction, cn_map and csm
            #  In site_ce_list["neighbors"], a dictionary with the corresponding neighbors (as a list of pymatgen's
            #    Site objects) for each cn_map is stored
            ce_piece_dict_list = []
            neighbors_list = {}
            for ce_piece in site_ce_list['ce']:
                if ce_piece['cn_map'] not in neighbors_list:
                    neighbors_list[ce_piece['cn_map']] = site_ce_list['neighbors'][ce_piece['cn_map']]
                ce_piece_dict = {'ce_symbol': ce_piece['mp_symbol'], 'fraction': ce_piece['fraction'],
                                 'csm': ce_piece['csm'], 'cn_map': ce_piece['cn_map']}
                ce_piece_dict_list.append(ce_piece_dict)
            self._coordination_environments.append(ce_piece_dict_list)
            self._neighbors.append(neighbors_list)
        if setup_neighbors_by_indices:
            self._setup_neighbors_by_indices()

    def _setup_neighbors_by_indices(self):
        if not self._strategy.uniquely_determines_coordination_environments:
            raise ValueError('Should use a strategy that uniquely determines coordination environments !')
        self._neighbors_by_indices = []
        for isite_this_site, this_site_neighbors_all_cn_maps in enumerate(self._neighbors):
            if len(self._coordination_environments[isite_this_site]) == 0:
                self._neighbors_by_indices.append(None)
                continue
            cn_map = self._coordination_environments[isite_this_site][0]['cn_map']
            this_site_neighbors = this_site_neighbors_all_cn_maps[tuple(cn_map)]
            if this_site_neighbors is None:
                self._neighbors_by_indices.append(None)
            else:
                this_site_neighbors_indices = []
                for neighbor in this_site_neighbors:
                    ineighbor = None
                    for isite, site in enumerate(self.structure):
                        found = False
                        for tolerance in [1e-8, 1e-6, 1e-4, 1e-3]:
                            if site.is_periodic_image(neighbor, check_lattice=False, tolerance=tolerance):
                                ineighbor = isite
                                diff = neighbor.frac_coords - site.frac_coords
                                image_cell = np.array(np.round(diff), np.int)
                                #if np.allclose(image_cell, np.zeros(3, np.int)):
                                #    image_cell = 0
                                found = True
                                break
                        if found:
                            break
                    if not found:
                        raise ChemenvError('LightStructureEnvironments', '__init__', 'Site indices not found')
                    this_site_neighbors_indices.append((ineighbor, image_cell))
                self._neighbors_by_indices.append(this_site_neighbors_indices)

    def setup_statistic_lists(self):
        self.statistics_dict = {'anion_list': {},                                                    # OK
                                'anion_number': None,                                                # OK
                                'anion_atom_list': {},                                               # OK
                                'anion_atom_number': None,                                           # OK
                                'cation_list': {},                                                   # OK
                                'cation_number': None,                                               # OK
                                'cation_atom_list': {},                                              # OK
                                'cation_atom_number': None,                                          # OK
                                'structure_anion_list': {},                                          # OK
                                'structure_anion_number': None,                                      # OK
                                'structure_anion_undefined_oxidation_list': [],                      # OK
                                'structure_anion_undefined_oxidation_occupation': [],                # OK
                                'structure_anion_atom_list': {},                                     # OK
                                'structure_anion_atom_number': None,                                 # OK
                                'structure_cation_list': {},                                         # OK
                                'structure_cation_number': None,                                     # OK
                                'structure_cation_undefined_oxidation_list': [],                     # OK
                                'structure_cation_undefined_oxidation_occupation': [],               # OK
                                'structure_cation_atom_list': {},                                    # OK
                                'structure_cation_atom_number': None,                                # OK
                                'bva_anion_list': {},                                                # OK
                                'bva_anion_number': None,                                            # OK
                                'bva_anion_atom_list': {},                                           # OK
                                'bva_anion_atom_number': None,                                       # OK
                                'bva_cation_list': {},                                               # OK
                                'bva_cation_number': None,                                           # OK
                                'bva_cation_atom_list': {},                                          # OK
                                'bva_cation_atom_number': None,                                      # OK
                                'neutral_list': {},                                                  # OK
                                'neutral_number': None,                                              # OK
                                'neutral_atom_list': {},                                             # OK
                                'neutral_atom_number': None,                                         # OK
                                'structure_neutral_list': {},                                        # OK
                                'structure_neutral_number': None,                                    # OK
                                'structure_neutral_atom_list': {},                                   # OK
                                'structure_neutral_atom_number': None,                               # OK
                                'bva_neutral_list': {},                                              # OK
                                'bva_neutral_number': None,                                          # OK
                                'bva_neutral_atom_list': {},                                         # OK
                                'bva_neutral_atom_number': None,                                     # OK
                                'atom_coordination_environments_present': {},                        # OK
                                'structure_ion_coordination_environments_present': {},               # OK
                                'bva_ion_coordination_environments_present': {},                     # OK
                                'ion_coordination_environments_present': {},                         # OK
                                'coordination_environments_structure_ion_present': {},               # OK
                                'coordination_environments_ion_present': {},                         # OK
                                'coordination_environments_bva_ion_present': {},                     # OK
                                'coordination_environments_atom_present': {},                        # OK
                                'fraction_bva_ion_coordination_environments_present': {},            # OK
                                'fraction_structure_ion_coordination_environments_present': {},      # OK
                                'fraction_ion_coordination_environments_present': {},                # OK
                                'fraction_atom_coordination_environments_present': {},               # OK
                                'fraction_coordination_environments_bva_ion_present': {},            # OK
                                'fraction_coordination_environments_structure_ion_present': {},      # OK
                                'fraction_coordination_environments_ion_present': {},                # OK
                                'fraction_coordination_environments_atom_present': {},               # OK
                                'count_bva_ion_present': {},                                         # OK
                                'count_structure_ion_present': {},                                   # OK
                                'count_ion_present': {},                                             # OK
                                'count_atom_present': {},                                            # OK
                                'count_coordination_environments_present': {}}
        structure_has_species = isinstance(self._structure[0].species_and_occu.elements[0], Specie)
        atom_stat = self.statistics_dict['atom_coordination_environments_present']
        ce_atom_stat = self.statistics_dict['coordination_environments_atom_present']
        fraction_atom_stat = self.statistics_dict['fraction_atom_coordination_environments_present']
        fraction_ce_atom_stat = self.statistics_dict['fraction_coordination_environments_atom_present']
        count_atoms = self.statistics_dict['count_atom_present']
        count_ce = self.statistics_dict['count_coordination_environments_present']
        for isite, site in enumerate(self._structure):
            #Building anion and cation list
            site_species = {'bva_oxistates': [],
                            'structure_oxistates': []}
            if self._bva_valences != 'undefined':
                for sp, occ in site.species_and_occu.items():
                    valence = self._bva_valences[isite]
                    if valence < 0:
                        specielist = self.statistics_dict['bva_anion_list']
                        atomlist = self.statistics_dict['bva_anion_atom_list']
                    elif valence > 0:
                        specielist = self.statistics_dict['bva_cation_list']
                        atomlist = self.statistics_dict['bva_cation_atom_list']
                    else:
                        specielist = self.statistics_dict['bva_neutral_list']
                        atomlist = self.statistics_dict['bva_neutral_atom_list']
                    strspecie = str(Specie(sp.symbol, valence))
                    if strspecie not in specielist:
                        specielist[strspecie] = occ
                    else:
                        specielist[strspecie] += occ
                    if sp.symbol not in atomlist:
                        atomlist[sp.symbol] = occ
                    else:
                        atomlist[sp.symbol] += occ
                    site_species['bva_oxistates'].append((sp.symbol, valence, occ))
                self.statistics_dict['bva_anion_number'] = len(self.statistics_dict['bva_anion_list'])
                self.statistics_dict['bva_anion_atom_number'] = len(self.statistics_dict['bva_anion_atom_list'])
                self.statistics_dict['bva_cation_number'] = len(self.statistics_dict['bva_cation_list'])
                self.statistics_dict['bva_cation_atom_number'] = len(self.statistics_dict['bva_cation_atom_list'])
            if structure_has_species:
                for sp, occ in site.species_and_occu.items():
                    oxi_state_rounded = round(sp.oxi_state)
                    if np.isclose(sp.oxi_state, oxi_state_rounded, rtol=0.0, atol=self.DELTA_MAX_OXIDATION_STATE):
                        valence = int(oxi_state_rounded)
                        if valence < 0:
                            specielist = self.statistics_dict['structure_anion_list']
                            atomlist = self.statistics_dict['structure_anion_atom_list']
                        elif valence > 0:
                            specielist = self.statistics_dict['structure_cation_list']
                            atomlist = self.statistics_dict['structure_cation_atom_list']
                        else:
                            specielist = self.statistics_dict['structure_neutral_list']
                            atomlist = self.statistics_dict['structure_neutral_atom_list']
                        strspecie = str(Specie(sp.symbol, valence))
                        if strspecie not in specielist:
                            specielist[strspecie] = occ
                        else:
                            specielist[strspecie] += occ
                        if sp.symbol not in atomlist:
                            atomlist[sp.symbol] = occ
                        else:
                            atomlist[sp.symbol] += occ
                        site_species['structure_oxistates'].append((sp.symbol, valence, occ))
                    else:
                        valence = sp.oxi_state
                        if valence < 0:
                            specielist = self.statistics_dict['structure_anion_undefined_oxidation_list']
                            specieocc = self.statistics_dict['structure_anion_undefined_oxidation_occupation']
                            atomlist = self.statistics_dict['structure_anion_atom_list']
                        elif valence > 0:
                            specielist = self.statistics_dict['structure_cation_undefined_oxidation_list']
                            specieocc = self.statistics_dict['structure_cation_undefined_oxidation_occupation']
                            atomlist = self.statistics_dict['structure_cation_atom_list']
                        else:
                            raise RuntimeError('Should not be here ...')
                        strspecie = str(Specie(sp.symbol, valence))
                        if strspecie not in specielist:
                            specielist.append(strspecie)
                            specieocc.append(occ)
                        else:
                            specie_index = specielist.index(strspecie)
                            specieocc[specie_index] += occ
                        if sp.symbol not in atomlist:
                            atomlist[sp.symbol] = occ
                        else:
                            atomlist[sp.symbol] += occ
                self.statistics_dict['structure_anion_number'] = len(self.statistics_dict['structure_anion_list']) + len(self.statistics_dict['structure_anion_undefined_oxidation_list'])
                self.statistics_dict['structure_anion_atom_number'] = len(self.statistics_dict['structure_anion_atom_list'])
                self.statistics_dict['structure_cation_number'] = len(self.statistics_dict['structure_cation_list']) + len(self.statistics_dict['structure_cation_undefined_oxidation_list'])
                self.statistics_dict['structure_cation_atom_number'] = len(self.statistics_dict['structure_cation_atom_list'])
            # for prefix in ['structure_', 'bva_']:
            #     if ((prefix == 'bva_' and self._bva_valences != 'undefined') or
            #             (prefix == 'structure_' and structure_has_species)):
            #         for sp, occ in site.species_and_occu.items():
            #             valence = self._bva_valences[isite] if prefix == 'bva_' else sp.oxi_state
            #             if valence < 0:
            #                 specielist = self.statistics_dict['{}anion_list'.format(prefix)]
            #                 atomlist = self.statistics_dict['{}anion_atom_list'.format(prefix)]
            #             elif valence > 0:
            #                 specielist = self.statistics_dict['{}cation_list'.format(prefix)]
            #                 atomlist = self.statistics_dict['{}cation_atom_list'.format(prefix)]
            #             else:
            #                 specielist = self.statistics_dict['{}neutral_list'.format(prefix)]
            #                 atomlist = self.statistics_dict['{}neutral_atom_list'.format(prefix)]
            #             strspecie = str(Specie(sp.symbol, valence))
            #             if strspecie not in specielist:
            #                 specielist[strspecie] = occ
            #             else:
            #                 specielist[strspecie] += occ
            #             if sp.symbol not in atomlist:
            #                 atomlist[sp.symbol] = occ
            #             else:
            #                 atomlist[sp.symbol] += occ
            #             site_species['{}oxistates'.format(prefix)].append((sp.symbol, valence, occ))
            #         self.statistics_dict['{}anion_number'.format(prefix)] = len(self.statistics_dict['{}anion_list'.format(prefix)])
            #         self.statistics_dict['{}anion_atom_number'.format(prefix)] = len(self.statistics_dict['{}anion_atom_list'.format(prefix)])
            #         self.statistics_dict['{}cation_number'.format(prefix)] = len(self.statistics_dict['{}cation_list'.format(prefix)])
            #         self.statistics_dict['{}cation_atom_number'.format(prefix)] = len(self.statistics_dict['{}cation_atom_list'.format(prefix)])

            #Building environments lists
            if self._coordination_environments[isite] is not None:
                site_envs = [(ce_piece_dict['ce_symbol'], ce_piece_dict['fraction'])
                             for ce_piece_dict in self._coordination_environments[isite]]
                for ce_symbol, fraction in site_envs:
                    if fraction is None:
                        continue
                    if ce_symbol not in count_ce:
                        count_ce[ce_symbol] = 0.0
                    count_ce[ce_symbol] += fraction
                for sp, occ in site.species_and_occu.items():
                    elmt = sp.symbol
                    if elmt not in atom_stat:
                        atom_stat[elmt] = {}
                        count_atoms[elmt] = 0.0
                    count_atoms[elmt] += occ
                    for ce_symbol, fraction in site_envs:
                        if fraction is None:
                            continue
                        if ce_symbol not in atom_stat[elmt]:
                            atom_stat[elmt][ce_symbol] = 0.0

                        atom_stat[elmt][ce_symbol] += occ * fraction
                        if ce_symbol not in ce_atom_stat:
                            ce_atom_stat[ce_symbol] = {}
                        if elmt not in ce_atom_stat[ce_symbol]:
                            ce_atom_stat[ce_symbol][elmt] = 0.0
                        ce_atom_stat[ce_symbol][elmt] += occ * fraction

                if self._bva_valences != 'undefined':
                    ion_stat = self.statistics_dict['bva_ion_coordination_environments_present']
                    ce_ion_stat = self.statistics_dict['coordination_environments_bva_ion_present']
                    count_ions = self.statistics_dict['count_bva_ion_present']
                    for elmt, oxi_state, occ in site_species['bva_oxistates']:
                        if elmt not in ion_stat:
                            ion_stat[elmt] = {}
                            count_ions[elmt] = {}
                        if oxi_state not in ion_stat[elmt]:
                            ion_stat[elmt][oxi_state] = {}
                            count_ions[elmt][oxi_state] = 0.0
                        count_ions[elmt][oxi_state] += occ
                        for ce_symbol, fraction in site_envs:
                            if fraction is None:
                                continue
                            if ce_symbol not in ion_stat[elmt][oxi_state]:
                                ion_stat[elmt][oxi_state][ce_symbol] = 0.0
                            ion_stat[elmt][oxi_state][ce_symbol] += occ * fraction
                            if ce_symbol not in ce_ion_stat:
                                ce_ion_stat[ce_symbol] = {}
                            if elmt not in ce_ion_stat[ce_symbol]:
                                ce_ion_stat[ce_symbol][elmt] = {}
                            if oxi_state not in ce_ion_stat[ce_symbol][elmt]:
                                ce_ion_stat[ce_symbol][elmt][oxi_state] = 0.0
                            ce_ion_stat[ce_symbol][elmt][oxi_state] += occ * fraction
                if structure_has_species:
                    ion_stat = self.statistics_dict['structure_ion_coordination_environments_present']
                    ce_ion_stat = self.statistics_dict['coordination_environments_structure_ion_present']
                    count_ions = self.statistics_dict['count_structure_ion_present']
                    for elmt, oxi_state, occ in site_species['structure_oxistates']:
                        if elmt not in ion_stat:
                            ion_stat[elmt] = {}
                            count_ions[elmt] = {}
                        oxi_state_rounded = round(oxi_state)
                        if np.isclose(oxi_state, oxi_state_rounded, rtol=0.0, atol=self.DELTA_MAX_OXIDATION_STATE):
                            my_oxi_state = int(oxi_state_rounded)
                            if my_oxi_state not in ion_stat[elmt]:
                                ion_stat[elmt][my_oxi_state] = {}
                                count_ions[elmt][my_oxi_state] = 0.0
                            count_ions[elmt][my_oxi_state] += occ
                            for ce_symbol, fraction in site_envs:
                                if fraction is None:
                                    continue
                                if ce_symbol not in ion_stat[elmt][my_oxi_state]:
                                    ion_stat[elmt][my_oxi_state][ce_symbol] = 0.0
                                ion_stat[elmt][my_oxi_state][ce_symbol] += occ * fraction
                                if ce_symbol not in ce_ion_stat:
                                    ce_ion_stat[ce_symbol] = {}
                                if elmt not in ce_ion_stat[ce_symbol]:
                                    ce_ion_stat[ce_symbol][elmt] = {}
                                if my_oxi_state not in ce_ion_stat[ce_symbol][elmt]:
                                    ce_ion_stat[ce_symbol][elmt][my_oxi_state] = 0.0
                                ce_ion_stat[ce_symbol][elmt][my_oxi_state] += occ * fraction
                        else:
                            #TODO: make something slightly better here ... like put the fractional oxidation states
                            #somewhere ...
                            continue
                # for prefix in ['structure_', 'bva_']:
                #     if prefix == 'bva_' and self._bva_valences == 'undefined':
                #         continue
                #     elif prefix == 'structure_' and not structure_has_species:
                #         continue
                #     ion_stat = self.statistics_dict['{}ion_coordination_environments_present'.format(prefix)]
                #     ce_ion_stat = self.statistics_dict['coordination_environments_{}ion_present'.format(prefix)]
                #     count_ions = self.statistics_dict['count_{}ion_present'.format(prefix)]
                #     for elmt, oxi_state, occ in site_species['{}oxistates'.format(prefix)]:
                #         if elmt not in ion_stat:
                #             ion_stat[elmt] = {}
                #             count_ions[elmt] = {}
                #         if oxi_state not in ion_stat[elmt]:
                #             ion_stat[elmt][oxi_state] = {}
                #             count_ions[elmt][oxi_state] = 0.0
                #         count_ions[elmt][oxi_state] += occ
                #         for ce_symbol, fraction in site_envs:
                #             if fraction is None:
                #                 continue
                #             if ce_symbol not in ion_stat[elmt][oxi_state]:
                #                 ion_stat[elmt][oxi_state][ce_symbol] = 0.0
                #             ion_stat[elmt][oxi_state][ce_symbol] += occ * fraction
                #             if ce_symbol not in ce_ion_stat:
                #                 ce_ion_stat[ce_symbol] = {}
                #             if elmt not in ce_ion_stat[ce_symbol]:
                #                 ce_ion_stat[ce_symbol][elmt] = {}
                #             if oxi_state not in ce_ion_stat[ce_symbol][elmt]:
                #                 ce_ion_stat[ce_symbol][elmt][oxi_state] = 0.0
                #             ce_ion_stat[ce_symbol][elmt][oxi_state] += occ * fraction
        for elmt, envs in atom_stat.items():
            sumelement = count_atoms[elmt]
            fraction_atom_stat[elmt] = {env: fraction / sumelement for env, fraction in envs.items()}
        for ce_symbol, atoms in ce_atom_stat.items():
            sumsymbol = count_ce[ce_symbol]
            fraction_ce_atom_stat[ce_symbol] = {atom: fraction / sumsymbol for atom, fraction in atoms.items()}
        for prefix in ['structure_', 'bva_']:
            ion_stat = self.statistics_dict['{}ion_coordination_environments_present'.format(prefix)]
            fraction_ion_stat = self.statistics_dict['fraction_{}ion_coordination_environments_present'.format(prefix)]
            ce_ion_stat = self.statistics_dict['coordination_environments_{}ion_present'.format(prefix)]
            fraction_ce_ion_stat = self.statistics_dict['fraction_coordination_environments_{}ion_present'.format(prefix)]
            count_ions = self.statistics_dict['count_{}ion_present'.format(prefix)]
            if prefix == 'bva_' and self._bva_valences == 'undefined':
                continue
            elif prefix == 'structure_' and not structure_has_species:
                continue
            for elmt, oxi_states_envs in ion_stat.items():
                fraction_ion_stat[elmt] = {}
                for oxi_state, envs in oxi_states_envs.items():
                    sumspecie = count_ions[elmt][oxi_state]
                    fraction_ion_stat[elmt][oxi_state] = {env: fraction / sumspecie
                                                          for env, fraction in envs.items()}
            for ce_symbol, ions in ce_ion_stat.items():
                fraction_ce_ion_stat[ce_symbol] = {}
                sum_ce = np.sum([np.sum(list(oxistates.values())) for elmt, oxistates in ions.items()])
                for elmt, oxistates in ions.items():
                    fraction_ce_ion_stat[ce_symbol][elmt] = {oxistate: fraction / sum_ce
                                                             for oxistate, fraction in oxistates.items()}
        ceip = 'coordination_environments_ion_present'
        icep = 'ion_coordination_environments_present'
        fceip = 'fraction_coordination_environments_ion_present'
        ficep = 'fraction_ion_coordination_environments_present'
        if self._bva_valences != 'undefined':
            cebip = 'coordination_environments_bva_ion_present'
            bicep = 'bva_ion_coordination_environments_present'
            fcebip = 'fraction_coordination_environments_bva_ion_present'
            fbicep = 'fraction_bva_ion_coordination_environments_present'
            self.statistics_dict['anion_list'].update(self.statistics_dict['bva_anion_list'])
            self.statistics_dict['anion_atom_list'].update(self.statistics_dict['bva_anion_atom_list'])
            self.statistics_dict['cation_list'].update(self.statistics_dict['bva_cation_list'])
            self.statistics_dict['cation_atom_list'].update(self.statistics_dict['bva_cation_atom_list'])
            self.statistics_dict['anion_number'] = self.statistics_dict['bva_anion_number']
            self.statistics_dict['anion_atom_number'] = self.statistics_dict['bva_anion_atom_number']
            self.statistics_dict['cation_number'] = self.statistics_dict['bva_cation_number']
            self.statistics_dict['cation_atom_number'] = self.statistics_dict['bva_cation_atom_number']
            self.statistics_dict[ceip].update(self.statistics_dict[cebip])
            self.statistics_dict[icep].update(self.statistics_dict[bicep])
            self.statistics_dict[fceip].update(self.statistics_dict[fcebip])
            self.statistics_dict[ficep].update(self.statistics_dict[fbicep])
        else:
            cesip = 'coordination_environments_structure_ion_present'
            sicep = 'structure_ion_coordination_environments_present'
            fcesip = 'fraction_coordination_environments_structure_ion_present'
            fsicep = 'fraction_structure_ion_coordination_environments_present'
            self.statistics_dict['anion_list'].update(self.statistics_dict['structure_anion_list'])
            self.statistics_dict['anion_atom_list'].update(self.statistics_dict['structure_anion_atom_list'])
            self.statistics_dict['cation_list'].update(self.statistics_dict['structure_cation_list'])
            self.statistics_dict['cation_atom_list'].update(self.statistics_dict['structure_cation_atom_list'])
            self.statistics_dict['anion_number'] = self.statistics_dict['structure_anion_number']
            self.statistics_dict['anion_atom_number'] = self.statistics_dict['structure_anion_atom_number']
            self.statistics_dict['cation_number'] = self.statistics_dict['structure_cation_number']
            self.statistics_dict['cation_atom_number'] = self.statistics_dict['structure_cation_atom_number']
            self.statistics_dict[ceip].update(self.statistics_dict[cesip])
            self.statistics_dict[icep].update(self.statistics_dict[sicep])
            self.statistics_dict[fceip].update(self.statistics_dict[fcesip])
            self.statistics_dict[ficep].update(self.statistics_dict[fsicep])

    def site_has_clear_environment(self, isite, min_fraction=0.95):
        if len(self._coordination_environments[isite]) == 0:
            return True
        fractions = [ce_dict['fraction'] for ce_dict in self._coordination_environments[isite]]
        maxfraction = max(fractions)
        return maxfraction > min_fraction

    def structure_has_clear_environments_for_element(self, element, min_fraction=0.95):
        for isite, site_ces in enumerate(self._coordination_environments):
            if site_ces is None:
                continue
            if element in [sp.symbol for sp in self._structure[isite].species_and_occu]:
                if not self.site_has_clear_environment(isite=isite, min_fraction=min_fraction):
                    return False
        return True

    def structure_has_clear_environments(self, min_fraction=0.95):
        for isite, site_ces in enumerate(self._coordination_environments):
            if site_ces is None:
                continue
            if not self.site_has_clear_environment(isite=isite, min_fraction=min_fraction):
                return False
        return True

    def get_site_info_for_specie_ce(self, specie, ce_symbol, min_fraction=0.0):
        element = specie.symbol
        oxi_state = specie.oxi_state
        isites = []
        csms = []
        fractions = []
        for isite, site in enumerate(self._structure):
            if element in [sp.symbol for sp in site.species_and_occu]:
                if oxi_state == self._bva_valences[isite]:
                    for ce_dict in self._coordination_environments[isite]:
                        if ce_symbol == ce_dict['ce_symbol']:
                            isites.append(isite)
                            csms.append(ce_dict['ce_symbol'])
                            fractions.append(ce_dict['fraction'])

    def get_site_info_for_specie_allces(self, specie, min_fraction=0.0):
        allces = {}
        element = specie.symbol
        oxi_state = specie.oxi_state
        for isite, site in enumerate(self._structure):
            if element in [sp.symbol for sp in site.species_and_occu]:
                if oxi_state == self._bva_valences[isite]:
                    for ce_dict in self._coordination_environments[isite]:
                        if ce_dict['fraction'] < min_fraction:
                            continue
                        if ce_dict['ce_symbol'] not in allces:
                            allces[ce_dict['ce_symbol']] = {'isites': [], 'fractions': [], 'csms': []}
                        allces[ce_dict['ce_symbol']]['isites'].append(isite)
                        allces[ce_dict['ce_symbol']]['fractions'].append(ce_dict['fraction'])
                        allces[ce_dict['ce_symbol']]['csms'].append(ce_dict['csm'])
        return allces

    def get_statistics(self, statistics_fields=DEFAULT_STATISTICS_FIELDS, bson_compatible=False):
        if self.statistics_dict is None:
            self.setup_statistic_lists()
        if bson_compatible:
            dd = jsanitize({field: self.statistics_dict[field] for field in statistics_fields})
        else:
            dd = {field: self.statistics_dict[field] for field in statistics_fields}
        return dd

    def contains_only_one_anion_atom(self, anion_atom):
        return (len(self.statistics_dict['anion_atom_list']) == 1 and
                anion_atom in self.statistics_dict['anion_atom_list'])

    def contains_only_one_anion(self, anion):
        return len(self.statistics_dict['anion_list']) == 1 and anion in self.statistics_dict['anion_list']

    def site_contains_environment(self, isite, ce_symbol):
        return ce_symbol in [ce_dict['ce_symbol'] for ce_dict in self._coordination_environments[isite]]

    def structure_contains_atom_environment(self, atom_symbol, ce_symbol):
        """
        Checks whether the structure contains a given atom in a given environment
        :param atom_symbol: Symbol of the atom
        :param ce_symbol: Symbol of the coordination environment
        :return: True if the coordination environment is found, False otherwise
        """
        for isite, site in enumerate(self._structure):
            if (Element(atom_symbol) in site.species_and_occu.
                    element_composition and self.site_contains_environment(isite, ce_symbol)):
                return True
        return False

    @property
    def coordination_environments(self):
        """
        Coordination environments determined by the strategy
        """
        return self._coordination_environments

    @property
    def uniquely_determined_coordination_environments(self):
        """
        True if the coordination environments are uniquely determined.
        """
        return self._uniquely_determined_coordination_environments

    @property
    def neighbors(self):
        """
        Neighbors determined by the strategy
        """
        return self._neighbors

    @property
    def neighbors_by_indices(self):
        """
        Neighbors by indices and image cell determined by the strategy
        """
        if self._neighbors_by_indices is None:
            self._setup_neighbors_by_indices()
        return self._neighbors_by_indices

    @property
    def structure(self):
        """
        Structure
        :return: Structure
        """
        return self._structure

    def __eq__(self, other):
        """
        Equality method that checks if the LightStructureEnvironments object is equal to another
        LightStructureEnvironments object. Two LightStructureEnvironments objects are equal if the strategy used
        is the same, if the structure is the same, if the valences used in the strategies are the same, if the
        coordination environments and the neighbours determined by the strategy are the same
        :param other: LightStructureEnvironments object to compare with
        :return: True if both objects are equal, False otherwise
        """
        return (self._strategy == other._strategy and self._structure == other._structure and
                self._bva_valences == other._bva_valences and
                self._coordination_environments == other._coordination_environments and
                self._neighbors == other._neighbors)

    def as_dict(self):
        """
        Bson-serializable dict representation of the LightStructureEnvironments object.
        :return: Bson-serializable dict representation of the LightStructureEnvironments object.
        """
        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "strategy": self._strategy.as_dict(),
                "structure": self._structure.as_dict(),
                "bva_valences": self._bva_valences,
                "coordination_environments": self._coordination_environments,
                "neighbors": [{'{:d}_{:d}'.format(cn_map[0], cn_map[1]): [ps.as_dict() for ps in neighbors]
                               for cn_map, neighbors in self._neighbors[isite].items()}
                              for isite in range(len(self._structure))],
                "neighbors_by_indices": jsanitize([None if item is None else list(item)
                                                   for item in self.neighbors_by_indices])}

    @classmethod
    def from_dict(cls, d):
        """
        Reconstructs the LightStructureEnvironments object from a dict representation of the
        LightStructureEnvironments created using the as_dict method.
        :param d: dict representation of the LightStructureEnvironments object
        :return: LightStructureEnvironments object
        """
        dec = MontyDecoder()
        neighbors = [{(int(cnmap.split('_')[0]), int(cnmap.split('_')[1])): dec.process_decoded(cnmap_neighbours)
                      for cnmap, cnmap_neighbours in neighbs_site.items()} for neighbs_site in d['neighbors']]
        coordination_environments = [[{key: val if key != 'cn_map' else tuple(val) for key, val in item.items()}
                                      for item in ces_site] for ces_site in d['coordination_environments']]
        if 'neighbors_by_indices' in d:
            neighbors_by_indices = []
            for item in d['neighbors_by_indices']:
                if item is None:
                    neighbors_by_indices.append(None)
                else:
                    neighbors_by_indices.append([(nb[0], np.array(nb[1], np.int)) for nb in item])
            # neighbors_by_indices = [None if item is None else (item[0], np.array(item[1], np.int))
            #                         for item in d['neighbors_by_indices']]
        else:
            neighbors_by_indices = None
        return cls(dec.process_decoded(d['strategy']), structure_environments=None,
                   structure=dec.process_decoded(d['structure']),
                   bva_valences=d['bva_valences'],
                   coordination_environments=coordination_environments,
                   neighbors=neighbors,
                   neighbors_by_indices=neighbors_by_indices)


# class LightStructureEnvironmentsOld(MSONable):
#     """
#     Class used to store the chemical environments of a given structure obtained from a given ChemenvStrategy. Currently,
#     only strategies leading to the determination of a unique environment for each site is allowed
#     This class does not store all the information contained in the StructureEnvironments object, only the coordination
#     environment found
#     """
#
#     def __init__(self, strategy_used, structure_environments=None, structure=None, bva_valences=None,
#                  coordination_environments=None, neighbors=None):
#         """
#         Constructor for the LightStructureEnvironments object.
#         """
#         if not strategy_used.uniquely_determines_coordination_environments:
#             raise NotImplementedError('LightStructureEnvironments not yet implemented with complex strategies leading '
#                                       'to multiple environments of a same site ...')
#         self.strategy_used = strategy_used
#         self._uniquely_determined_coordination_environments = strategy_used.uniquely_determines_coordination_environments
#         self._neighbors = []
#         self._coordination_environments = []
#         if structure_environments is not None:
#             self.strategy_used.set_structure_environments(structure_environments)
#             self.structure = structure_environments.structure
#             self.bva_valences = structure_environments.bva_valences
#             for site in self.strategy_used.structure_environments.structure:
#                 try:
#                     this_site_neighbors = self.strategy_used.get_site_neighbors(site)
#                     self._neighbors.append(this_site_neighbors)
#                     try:
#                         (ce, ce_dict) = self.strategy_used.get_site_coordination_environment(site)
#                         self._coordination_environments.append(ce)
#                     except TypeError:
#                         cn = self.strategy_used.get_site_coordination_environment(site)
#                         self._coordination_environments.append(cn)
#                 except NeighborsNotComputedChemenvError:
#                     self._neighbors.append(None)
#                     self._coordination_environments.append(None)
#         else:
#             if structure is None or coordination_environments is None or neighbors is None or bva_valences is None:
#                 raise InitializationChemenvError(self.__class__.__name__)
#             self.structure = structure
#             self._coordination_environments = coordination_environments
#             self._neighbors = neighbors
#             self.bva_valences = bva_valences
#         self._neighbors_by_indices = []
#         for this_site_neighbors in self._neighbors:
#             if this_site_neighbors is None:
#                 self._neighbors_by_indices.append(None)
#             else:
#                 this_site_neighbors_indices = []
#                 for neighbor in this_site_neighbors:
#                     ineighbor = None
#                     for isite, site in enumerate(self.structure):
#                         found = False
#                         for tolerance in [1e-8, 1e-6, 1e-4, 1e-3]:
#                             if site.is_periodic_image(neighbor, check_lattice=False, tolerance=tolerance):
#                                 ineighbor = isite
#                                 diff = neighbor.frac_coords - site.frac_coords
#                                 image_cell = np.array(np.round(diff), np.int)
#                                 #if np.allclose(image_cell, np.zeros(3, np.int)):
#                                 #    image_cell = 0
#                                 found = True
#                                 break
#                         if found:
#                             break
#                     if not found:
#                         raise ChemenvError('LightStructureEnvironments', '__init__', 'Site indices not found')
#                     this_site_neighbors_indices.append((ineighbor, image_cell))
#                 self._neighbors_by_indices.append(this_site_neighbors_indices)
#         self.setup_statistic_lists()
#
#     def contains_only_one_anion_atom(self, anion_atom):
#         return len(self.anion_atom_list) == 1 and self.anion_atom_list[0] == anion_atom
#
#     def contains_only_one_anion(self, anion):
#         return len(self.anion_list) == 1 and self.anion_list[0] == anion
#
#     def site_contains_environment(self, isite, mp_symbol):
#         if self.uniquely_determined_coordination_environments:
#             return self.coordination_environments[isite] == mp_symbol
#         else:
#             return mp_symbol in self.coordination_environments[isite]
#
#     def structure_contains_atom_environment(self, atom_symbol, mp_symbol):
#         """
#         Checks whether the structure contains a given atom in a given environment
#         :param atom_symbol: Symbol of the atom
#         :param mp_symbol: Symbol of the coordination environment
#         :return: True if the coordination environment is found, False otherwise
#         """
#         if self.uniquely_determined_coordination_environments:
#             for isite, site in enumerate(self.structure):
#                 if Element(atom_symbol) in site.species_and_occu.element_composition:
#                     if self.coordination_environments[isite] == mp_symbol:
#                         return True
#             return False
#         else:
#             for isite, site in enumerate(self.structure):
#                 if Element(atom_symbol) in site.species_and_occu.element_composition:
#                     if mp_symbol in self.coordination_environments[isite]:
#                         return True
#             return False
#
#     def setup_statistic_lists(self):
#         self.anion_list = []
#         self.anion_atom_list = []
#         self.cation_list = []
#         self.cation_atom_list = []
#         self.coordination_environments_present = {}
#         self.site_count_with_computed_ce = 0.0
#         self.atom_coordination_environments_present = {}
#         self.atom_count_with_computed_ce = {}
#         self.structure_ion_coordination_environments_present = {}
#         self.structure_ion_count_with_computed_ce = {}
#         self.bva_ion_coordination_environments_present = {}
#         self.bva_ion_count_with_computed_ce = {}
#         self.coordination_environments_bva_ion_present = {}
#         self.fraction_bva_ion_coordination_environments_present = {}
#         self.fraction_coordination_environments_bva_ion_present = {}
#         for isite, site in enumerate(self.structure):
#             #Building anion list
#             if self.bva_valences != 'undefined':
#                 for sp, occ in site.species_and_occu.items():
#                     if self.bva_valences[isite] < 0:
#                         stranion = str(Specie(sp.symbol, self.bva_valences[isite]))
#                         if stranion not in self.anion_list:
#                             self.anion_list.append(stranion)
#                         if sp.symbol not in self.anion_atom_list:
#                             self.anion_atom_list.append(sp.symbol)
#                     else:
#                         strcation = str(Specie(sp.symbol, self.bva_valences[isite]))
#                         if strcation not in self.cation_list:
#                             self.cation_list.append(strcation)
#                         if sp.symbol not in self.cation_atom_list:
#                             self.cation_atom_list.append(sp.symbol)
#             if self.coordination_environments[isite] is not None:
#                 ce = '{}'.format(self.coordination_environments[isite])
#                 self.site_count_with_computed_ce += 1.0
#                 if not ce in self.coordination_environments_present:
#                     self.coordination_environments_present[ce] = 1
#                 else:
#                     self.coordination_environments_present[ce] += 1
#                 for specie in site.species_and_occu:
#                     #Counting atoms in specific environments
#                     if specie.symbol not in self.atom_coordination_environments_present:
#                         self.atom_coordination_environments_present[specie.symbol] = {}
#                     atom_ce = '{}'.format(self.coordination_environments[isite])
#                     if not atom_ce in self.atom_coordination_environments_present:
#                         self.atom_coordination_environments_present[specie.symbol][atom_ce] = site.species_and_occu[specie]
#                     else:
#                         self.atom_coordination_environments_present[specie.symbol][atom_ce] += site.species_and_occu[specie]
#                     #Counting atoms for which environments are determined
#                     if specie.symbol not in self.atom_count_with_computed_ce:
#                         self.atom_count_with_computed_ce[specie.symbol] = site.species_and_occu[specie]
#                     else:
#                         self.atom_count_with_computed_ce[specie.symbol] += site.species_and_occu[specie]
#                     #Counting ions in specific environments
#                     if specie.symbol not in self.structure_ion_coordination_environments_present:
#                         self.structure_ion_coordination_environments_present[specie.symbol] = {}
#                     try:
#                         str_sp_oxi_state = str(specie.oxi_state)
#                     except AttributeError:
#                         str_sp_oxi_state = 'element'
#                     str_sp_oxi_state = str_sp_oxi_state.replace('.', ',')
#                     if str_sp_oxi_state not in self.structure_ion_coordination_environments_present[specie.symbol]:
#                         self.structure_ion_coordination_environments_present[specie.symbol][str_sp_oxi_state] = {}
#                     ion_ce = '{}'.format(self.coordination_environments[isite])
#                     if ion_ce not in self.structure_ion_coordination_environments_present[specie.symbol][str_sp_oxi_state]:
#                         self.structure_ion_coordination_environments_present[specie.symbol][str_sp_oxi_state][ion_ce] = site.species_and_occu[specie]
#                     else:
#                         self.structure_ion_coordination_environments_present[specie.symbol][str_sp_oxi_state][ion_ce] += site.species_and_occu[specie]
#                     #Counting ions for which environments are determined
#                     if specie.symbol not in self.structure_ion_count_with_computed_ce:
#                         self.structure_ion_count_with_computed_ce[specie.symbol] = {}
#                     if str_sp_oxi_state not in self.structure_ion_count_with_computed_ce[specie.symbol]:
#                         self.structure_ion_count_with_computed_ce[specie.symbol][str_sp_oxi_state] = site.species_and_occu[specie]
#                     else:
#                         self.structure_ion_count_with_computed_ce[specie.symbol][str_sp_oxi_state] += site.species_and_occu[specie]
#                         #Counting ions in specific environments
#                     if specie.symbol not in self.bva_ion_coordination_environments_present:
#                         self.bva_ion_coordination_environments_present[specie.symbol] = {}
#                     if self.bva_valences != 'undefined':
#                         str_val = str(self.bva_valences[isite])
#                         if str_val not in self.bva_ion_coordination_environments_present[specie.symbol]:
#                             self.bva_ion_coordination_environments_present[specie.symbol][str_val] = {}
#                         ion_ce = '{}'.format(self.coordination_environments[isite])
#                         if ion_ce not in self.bva_ion_coordination_environments_present[specie.symbol][str_val]:
#                             self.bva_ion_coordination_environments_present[specie.symbol][str_val][ion_ce] = site.species_and_occu[specie]
#                         else:
#                             self.bva_ion_coordination_environments_present[specie.symbol][str_val][ion_ce] += site.species_and_occu[specie]
#                         if specie.symbol not in self.bva_ion_count_with_computed_ce:
#                             self.bva_ion_count_with_computed_ce[specie.symbol] = {}
#                         if str_val not in self.bva_ion_count_with_computed_ce[specie.symbol]:
#                             self.bva_ion_count_with_computed_ce[specie.symbol][str_val] = site.species_and_occu[specie]
#                         else:
#                             self.bva_ion_count_with_computed_ce[specie.symbol][str_val] += site.species_and_occu[specie]
#         for ce in self.coordination_environments_present:
#             self.coordination_environments_bva_ion_present[ce] = {}
#         for sp, spdict in self.bva_ion_coordination_environments_present.items():
#             self.fraction_bva_ion_coordination_environments_present[sp] = {}
#             for val, valdict in spdict.items():
#                 strbvasp = str(Specie(sp, int(val)))
#                 self.fraction_bva_ion_coordination_environments_present[sp][val] = {}
#                 for ion_ce, n_ion_ce in valdict.items():
#                     self.fraction_bva_ion_coordination_environments_present[sp][val][ion_ce] = float(n_ion_ce) / self.bva_ion_count_with_computed_ce[sp][val]
#                     if strbvasp in self.coordination_environments_bva_ion_present[ion_ce]:
#                         self.coordination_environments_bva_ion_present[ion_ce][strbvasp] += n_ion_ce
#                     else:
#                         self.coordination_environments_bva_ion_present[ion_ce][strbvasp] = n_ion_ce
#         for ce, cedict in self.coordination_environments_bva_ion_present.items():
#             self.fraction_coordination_environments_bva_ion_present[ce] = {}
#             for cation, ncations in cedict.items():
#                 self.fraction_coordination_environments_bva_ion_present[ce][cation] = float(ncations) / self.coordination_environments_present[ce]
#
#     @property
#     def coordination_environments(self):
#         return self._coordination_environments
#
#     @property
#     def uniquely_determined_coordination_environments(self):
#         return self._uniquely_determined_coordination_environments
#
#     @property
#     def neighbors(self):
#         return self._neighbors
#
#     @property
#     def neighbors_by_indices(self):
#         return self._neighbors_by_indices
#
#     def __eq__(self, other):
#         return (self.strategy_used == other.strategy_used and self.structure == other.structure and
#                 self.bva_valences == other.bva_valences and
#                 self._coordination_environments == other._coordination_environments and
#                 self._neighbors == other._neighbors)
#
#     def as_dict(self):
#         """
#         Bson-serializable dict representation of the LightStructureEnvironments object.
#         :return: Bson-serializable dict representation of the LightStructureEnvironments object.
#         """
#         #strategy_used, structure_environments=None, structure=None, coordination_environments=None,
#         #         neighbors=None
#         return {"@module": self.__class__.__module__,
#                 "@class": self.__class__.__name__,
#                 "strategy_used": self.strategy_used.as_dict(),
#                 "structure": self.structure.as_dict(),
#                 "bva_valences": self.bva_valences,
#                 "coordination_environments": self._coordination_environments,
#                 "neighbors": [[ps.as_dict() for ps in neighbs] if neighbs is not None else None
#                               for neighbs in self._neighbors]}
#
#     @classmethod
#     def from_dict(cls, d):
#         """
#         Reconstructs the LightStructureEnvironments object from a dict representation of the
#         LightStructureEnvironments created using the as_dict method.
#         :param d: dict representation of the LightStructureEnvironments object
#         :return: LightStructureEnvironments object
#         """
#         dec = MontyDecoder()
#         return cls(dec.process_decoded(d['strategy_used']), structure_environments=None,
#                    structure=dec.process_decoded(d['structure']),
#                    bva_valences=d['bva_valences'],
#                    coordination_environments=d['coordination_environments'],
#                    neighbors=dec.process_decoded(d['neighbors']))


class ChemicalEnvironments(MSONable):
    """
    Class used to store all the information about the chemical environment of a given site for a given list of
    coordinated neighbours (internally called "cn_map")
    """

    def __init__(self, coord_geoms=None):
        """
        Initializes the ChemicalEnvironments object containing all the information about the chemical
        environment of a given site
        :param coord_geoms: coordination geometries to be added to the chemical environment.
        """
        if coord_geoms is None:
            self.coord_geoms = {}
        else:
            raise NotImplementedError('Constructor for ChemicalEnvironments with the coord_geoms argument is not'
                                      'yet implemented')

    def __getitem__(self, mp_symbol):
        if not mp_symbol in self.coord_geoms:
            raise IndexError()
        return self.coord_geoms[mp_symbol]

    def __len__(self):
        """
        Returns the number of coordination geometries in this ChemicalEnvironments object
        :return: Number of coordination geometries in this ChemicalEnvironments object
        """
        return len(self.coord_geoms)

    def minimum_geometry(self, symmetry_measure_type=None, max_csm=None):
        """
        Returns the geometry with the minimum continuous symmetry measure of this ChemicalEnvironments
        :return: tuple (symbol, csm) with symbol being the geometry with the minimum continuous symmetry measure and
        csm being the continuous symmetry measure associted to it
        :raise: ValueError if no coordination geometry is found in this ChemicalEnvironments object
        """
        if len(self.coord_geoms) == 0:
            return None
        cglist = [cg for cg in self.coord_geoms]
        if symmetry_measure_type is None:
            csms = np.array([self.coord_geoms[cg]['other_symmetry_measures']['csm_wcs_ctwcc'] for cg in cglist])
        else:
            csms = np.array([self.coord_geoms[cg]['other_symmetry_measures'][symmetry_measure_type] for cg in cglist])
        csmlist = [self.coord_geoms[cg] for cg in cglist]
        imin = np.argmin(csms)
        if max_csm is not None:
            if csmlist[imin] > max_csm:
                return None
        return cglist[imin], csmlist[imin]

    def minimum_geometries(self, n=None, symmetry_measure_type=None, max_csm=None):
        """
        Returns a list of geometries with increasing continuous symmetry measure in this ChemicalEnvironments object
        :param n: Number of geometries to be included in the list
        :return: list of geometries with increasing continuous symmetry measure in this ChemicalEnvironments object
        :raise: ValueError if no coordination geometry is found in this ChemicalEnvironments object
        """
        cglist = [cg for cg in self.coord_geoms]
        if symmetry_measure_type is None:
            csms = np.array([self.coord_geoms[cg]['other_symmetry_measures']['csm_wcs_ctwcc'] for cg in cglist])
        else:
            csms = np.array([self.coord_geoms[cg]['other_symmetry_measures'][symmetry_measure_type] for cg in cglist])
        csmlist = [self.coord_geoms[cg] for cg in cglist]
        isorted = np.argsort(csms)
        if max_csm is not None:
            if n is None:
                return [(cglist[ii], csmlist[ii]) for ii in isorted if csmlist[ii] <= max_csm]
            else:
                return [(cglist[ii], csmlist[ii]) for ii in isorted[:n] if csmlist[ii] <= max_csm]
        else:
            if n is None:
                return [(cglist[ii], csmlist[ii]) for ii in isorted]
            else:
                return [(cglist[ii], csmlist[ii]) for ii in isorted[:n]]

    def add_coord_geom(self, mp_symbol, symmetry_measure, algo='UNKNOWN', permutation=None, override=False,
                       local2perfect_map=None, perfect2local_map=None, detailed_voronoi_index=None,
                       other_symmetry_measures=None):
        """
        Adds a coordination geometry to the ChemicalEnvironments object
        :param mp_symbol: Symbol (internal) of the coordination geometry added
        :param symmetry_measure: Symmetry measure of the coordination geometry added
        :param algo: Algorithm used for the search of the coordination geometry added
        :param permutation: Permutation of the neighbors that leads to the csm stored
        :param override: If set to True, the coordination geometry will override the existent one if present
        :return: :raise: ChemenvError if the coordination geometry is already added and override is set to False
        """
        if not allcg.is_a_valid_coordination_geometry(mp_symbol=mp_symbol):
            raise ChemenvError(self.__class__,
                               'add_coord_geom',
                               'Coordination geometry with mp_symbol "{mp}" is not valid'
                               .format(mp=mp_symbol))
        if mp_symbol in list(self.coord_geoms.keys()) and not override:
            raise ChemenvError(self.__class__,
                               "add_coord_geom",
                               "This coordination geometry is already present and override is set to False")
        else:
            self.coord_geoms[mp_symbol] = {'symmetry_measure': float(symmetry_measure), 'algo': algo,
                                           'permutation': [int(i) for i in permutation],
                                           'local2perfect_map': local2perfect_map,
                                           'perfect2local_map': perfect2local_map,
                                           'detailed_voronoi_index': detailed_voronoi_index,
                                           'other_symmetry_measures': other_symmetry_measures}

    def __str__(self):
        """
        Returns a string representation of the ChemicalEnvironments object
        :return: String representation of the ChemicalEnvironments object
        """
        out = 'Chemical environments object :\n'
        if len(self.coord_geoms) == 0:
            out += ' => No coordination in it <=\n'
            return out
        for key in self.coord_geoms.keys():
            mp_symbol = key
            break
        cn = symbol_cn_mapping[mp_symbol]
        out += ' => Coordination {} <=\n'.format(cn)
        mp_symbols = list(self.coord_geoms.keys())
        csms_wcs = [self.coord_geoms[mp_symbol]['other_symmetry_measures']['csm_wcs_ctwcc'] for mp_symbol in mp_symbols]
        icsms_sorted = np.argsort(csms_wcs)
        mp_symbols = [mp_symbols[ii] for ii in icsms_sorted]
        for mp_symbol in mp_symbols:
            csm_wcs = self.coord_geoms[mp_symbol]['other_symmetry_measures']['csm_wcs_ctwcc']
            csm_wocs = self.coord_geoms[mp_symbol]['other_symmetry_measures']['csm_wocs_ctwocc']
            out += '   - {}\n'.format(mp_symbol)
            out += '      csm1 (with central site) : {}'.format(csm_wcs)
            out += '      csm2 (without central site) : {}'.format(csm_wocs)
            out += '     algo : {}'.format(self.coord_geoms[mp_symbol]['algo'])
            out += '     perm : {}\n'.format(self.coord_geoms[mp_symbol]['permutation'])
            out += '       local2perfect : {}\n'.format(str(self.coord_geoms[mp_symbol]['local2perfect_map']))
            out += '       perfect2local : {}\n'.format(str(self.coord_geoms[mp_symbol]['perfect2local_map']))
        return out

    def __eq__(self, other):
        """
        Equality method that checks if the ChemicalEnvironments object is equal to another ChemicalEnvironments
        object.
        :param other: ChemicalEnvironments object to compare with
        :return: True if both objects are equal, False otherwise
        """
        return self.coord_geoms == other.coord_geoms

    def as_dict(self):
        """
        Returns a dictionary representation of the ChemicalEnvironments object
        :return:
        """
        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "coord_geoms": jsanitize(self.coord_geoms)}

    @classmethod
    def from_dict(cls, d):
        """
        Reconstructs the ChemicalEnvironments object from a dict representation of the ChemicalEnvironments created
        using the as_dict method.
        :param d: dict representation of the ChemicalEnvironments object
        :return: ChemicalEnvironments object
        """
        ce = cls()
        for cg in d['coord_geoms'].keys():
            if d['coord_geoms'][cg]['local2perfect_map'] is None:
                l2p_map = None
            else:
                l2p_map = {int(key): int(val) for key, val in d['coord_geoms'][cg]['local2perfect_map'].items()}
            if d['coord_geoms'][cg]['perfect2local_map'] is None:
                p2l_map = None
            else:
                p2l_map = {int(key): int(val) for key, val in d['coord_geoms'][cg]['perfect2local_map'].items()}
            if ('other_symmetry_measures' in d['coord_geoms'][cg] and
                        d['coord_geoms'][cg]['other_symmetry_measures'] is not None):
                other_csms = d['coord_geoms'][cg]['other_symmetry_measures']
            else:
                other_csms = None
            ce.add_coord_geom(cg,
                              d['coord_geoms'][cg]['symmetry_measure'],
                              d['coord_geoms'][cg]['algo'],
                              permutation=d['coord_geoms'][cg]['permutation'],
                              local2perfect_map=l2p_map,
                              perfect2local_map=p2l_map,
                              detailed_voronoi_index=d['coord_geoms'][cg]['detailed_voronoi_index'],
                              other_symmetry_measures=other_csms)
        return ce



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals

"""
Defines SymmetryGroup parent class and PointGroup and SpaceGroup classes.
Shyue Ping Ong thanks Marc De Graef for his generous sharing of his
SpaceGroup data as published in his textbook "Structure of Materials".
"""

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2013, The Materials Virtual Lab"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "ongsp@ucsd.edu"
__date__ = "4/4/14"

import os
from itertools import product
from fractions import Fraction
from abc import ABCMeta, abstractproperty
from collections import Sequence
import numpy as np
import warnings
from monty.serialization import loadfn

from pymatgen.core.operations import SymmOp
from monty.design_patterns import cached_class

SYMM_DATA = loadfn(os.path.join(os.path.dirname(__file__), "symm_data.yaml"))


GENERATOR_MATRICES = SYMM_DATA["generator_matrices"]
POINT_GROUP_ENC = SYMM_DATA["point_group_encoding"]
SPACE_GROUP_ENC = SYMM_DATA["space_group_encoding"]
ABBREV_SPACE_GROUP_MAPPING = SYMM_DATA["abbreviated_spacegroup_symbols"]
TRANSLATIONS = {k: Fraction(v) for k, v in SYMM_DATA["translations"].items()}
FULL_SPACE_GROUP_MAPPING = {
    v["full_symbol"]: k for k, v in SYMM_DATA["space_group_encoding"].items()}
MAXIMAL_SUBGROUPS = {int(k): v
                     for k, v in SYMM_DATA["maximal_subgroups"].items()}


class SymmetryGroup(Sequence):
    __metaclass__ = ABCMeta

    @abstractproperty
    def symmetry_ops(self):
        pass

    def __contains__(self, item):
        for i in self.symmetry_ops:
            if np.allclose(i.affine_matrix, item.affine_matrix):
                return True
        return False

    def __hash__(self):
        return self.__len__()

    def __getitem__(self, item):
        return self.symmetry_ops[item]

    def __len__(self):
        return len(self.symmetry_ops)

    def is_subgroup(self, supergroup):
        """
        True if this group is a subgroup of the supplied group.

        Args:
            supergroup (SymmetryGroup): Supergroup to test.

        Returns:
            True if this group is a subgroup of the supplied group.
        """
        warnings.warn("This is not fully functional. Only trivial subsets are tested right now. ")
        return set(self.symmetry_ops).issubset(supergroup.symmetry_ops)

    def is_supergroup(self, subgroup):
        """
        True if this group is a supergroup of the supplied group.

        Args:
            subgroup (SymmetryGroup): Subgroup to test.

        Returns:
            True if this group is a supergroup of the supplied group.
        """
        warnings.warn("This is not fully functional. Only trivial subsets are tested right now. ")
        return set(subgroup.symmetry_ops).issubset(self.symmetry_ops)


@cached_class
class PointGroup(SymmetryGroup):
    """
    Class representing a Point Group, with generators and symmetry operations.

    .. attribute:: symbol

        Full International or Hermann-Mauguin Symbol.

    .. attribute:: generators

        List of generator matrices. Note that 3x3 matrices are used for Point
        Groups.

    .. attribute:: symmetry_ops

        Full set of symmetry operations as matrices.
    """

    def __init__(self, int_symbol):
        """
        Initializes a Point Group from its international symbol.

        Args:
            int_symbol (str): International or Hermann-Mauguin Symbol.
        """
        self.symbol = int_symbol
        self.generators = [GENERATOR_MATRICES[c]
                           for c in POINT_GROUP_ENC[int_symbol]]
        self._symmetry_ops = set([SymmOp.from_rotation_and_translation(m)
                                  for m in self._generate_full_symmetry_ops()])
        self.order = len(self._symmetry_ops)

    @property
    def symmetry_ops(self):
        return self._symmetry_ops

    def _generate_full_symmetry_ops(self):
        symm_ops = list(self.generators)
        new_ops = self.generators
        while len(new_ops) > 0:
            gen_ops = []
            for g1, g2 in product(new_ops, symm_ops):
                op = np.dot(g1, g2)
                if not in_array_list(symm_ops, op):
                    gen_ops.append(op)
                    symm_ops.append(op)
            new_ops = gen_ops
        return symm_ops

    def get_orbit(self, p, tol=1e-5):
        """
        Returns the orbit for a point.

        Args:
            p: Point as a 3x1 array.
            tol: Tolerance for determining if sites are the same. 1e-5 should
                be sufficient for most purposes. Set to 0 for exact matching
                (and also needed for symbolic orbits).

        Returns:
            ([array]) Orbit for point.
        """
        orbit = []
        for o in self.symmetry_ops:
            pp = o.operate(p)
            if not in_array_list(orbit, pp, tol=tol):
                orbit.append(pp)
        return orbit


@cached_class
class SpaceGroup(SymmetryGroup):
    """
    Class representing a SpaceGroup.

    .. attribute:: symbol

        Full International or Hermann-Mauguin Symbol.

    .. attribute:: int_number

        International number

    .. attribute:: generators

        List of generator matrices. Note that 4x4 matrices are used for Space
        Groups.

    .. attribute:: order

        Order of Space Group
    """

    # Contains the entire list of supported Space Group symbols.
    SG_SYMBOLS = tuple(SPACE_GROUP_ENC.keys())

    def __init__(self, int_symbol):
        """
        Initializes a Space Group from its full or abbreviated international
        symbol. Only standard settings are supported.

        Args:
            int_symbol (str): Full International (e.g., "P2/m2/m2/m") or
                Hermann-Mauguin Symbol ("Pmmm") or abbreviated symbol. The
                notation is a LaTeX-like string, with screw axes being
                represented by an underscore. For example, "P6_3/mmc". Note
                that for rhomohedral cells, the hexagonal setting can be
                accessed by adding a "H", e.g., "R-3mH".
        """
        if int_symbol not in SPACE_GROUP_ENC and int_symbol not in \
                ABBREV_SPACE_GROUP_MAPPING and int_symbol not in \
                FULL_SPACE_GROUP_MAPPING:
            raise ValueError("Bad international symbol %s" % int_symbol)
        elif int_symbol in ABBREV_SPACE_GROUP_MAPPING:
            int_symbol = ABBREV_SPACE_GROUP_MAPPING[int_symbol]
        elif int_symbol in FULL_SPACE_GROUP_MAPPING:
            int_symbol = FULL_SPACE_GROUP_MAPPING[int_symbol]

        data = SPACE_GROUP_ENC[int_symbol]

        self.symbol = int_symbol
        # TODO: Support different origin choices.
        enc = list(data["enc"])
        inversion = int(enc.pop(0))
        ngen = int(enc.pop(0))
        symm_ops = [np.eye(4)]
        if inversion:
            symm_ops.append(np.array(
                [[-1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0],
                 [0, 0, 0, 1]]))
        for i in range(ngen):
            m = np.eye(4)
            m[:3, :3] = GENERATOR_MATRICES[enc.pop(0)]
            m[0, 3] = TRANSLATIONS[enc.pop(0)]
            m[1, 3] = TRANSLATIONS[enc.pop(0)]
            m[2, 3] = TRANSLATIONS[enc.pop(0)]
            symm_ops.append(m)
        self.generators = symm_ops
        self.full_symbol = data["full_symbol"]
        self.int_number = data["int_number"]
        self.order = data["order"]
        self.patterson_symmetry = data["patterson_symmetry"]
        self.point_group = data["point_group"]
        self._symmetry_ops = None

    def _generate_full_symmetry_ops(self):
        symm_ops = np.array(self.generators)
        for op in symm_ops:
            op[0:3, 3] = np.mod(op[0:3, 3], 1)
        new_ops = symm_ops
        while len(new_ops) > 0 and len(symm_ops) < self.order:
            gen_ops = []
            for g in new_ops:
                temp_ops = np.einsum('ijk,kl', symm_ops, g)
                for op in temp_ops:
                    op[0:3, 3] = np.mod(op[0:3, 3], 1)
                    ind = np.where(np.abs(1 - op[0:3, 3]) < 1e-5)
                    op[ind, 3] = 0
                    if not in_array_list(symm_ops, op):
                        gen_ops.append(op)
                        symm_ops = np.append(symm_ops, [op], axis=0)
            new_ops = gen_ops
        assert len(symm_ops) == self.order
        return symm_ops

    @property
    def symmetry_ops(self):
        """
        Full set of symmetry operations as matrices. Lazily initialized as
        generation sometimes takes a bit of time.
        """
        if self._symmetry_ops is None:
            self._symmetry_ops = [
                SymmOp(m) for m in self._generate_full_symmetry_ops()]
        return self._symmetry_ops

    def get_orbit(self, p, tol=1e-5):
        """
        Returns the orbit for a point.

        Args:
            p: Point as a 3x1 array.
            tol: Tolerance for determining if sites are the same. 1e-5 should
                be sufficient for most purposes. Set to 0 for exact matching
                (and also needed for symbolic orbits).

        Returns:
            ([array]) Orbit for point.
        """
        orbit = []
        for o in self.symmetry_ops:
            pp = o.operate(p)
            pp = np.mod(np.round(pp, decimals=10), 1)
            if not in_array_list(orbit, pp, tol=tol):
                orbit.append(pp)
        return orbit

    def is_compatible(self, lattice, tol=1e-5, angle_tol=5):
        """
        Checks whether a particular lattice is compatible with the
        *conventional* unit cell.

        Args:
            lattice (Lattice): A Lattice.
            tol (float): The tolerance to check for equality of lengths.
            angle_tol (float): The tolerance to check for equality of angles
                in degrees.
        """
        abc, angles = lattice.lengths_and_angles
        crys_system = self.crystal_system

        def check(param, ref, tolerance):
            return all([abs(i - j) < tolerance for i, j in zip(param, ref)
                        if j is not None])

        if crys_system == "cubic":
            a = abc[0]
            return check(abc, [a, a, a], tol) and\
                check(angles, [90, 90, 90], angle_tol)
        elif crys_system == "hexagonal" or (crys_system == "trigonal" and
                                            self.symbol.endswith("H")):
            a = abc[0]
            return check(abc, [a, a, None], tol)\
                and check(angles, [90, 90, 120], angle_tol)
        elif crys_system == "trigonal":
            a = abc[0]
            return check(abc, [a, a, a], tol)
        elif crys_system == "tetragonal":
            a = abc[0]
            return check(abc, [a, a, None], tol) and\
                check(angles, [90, 90, 90], angle_tol)
        elif crys_system == "orthorhombic":
            return check(angles, [90, 90, 90], angle_tol)
        elif crys_system == "monoclinic":
            return check(angles, [90, None, 90], angle_tol)
        return True

    @property
    def crystal_system(self):
        i = self.int_number
        if i <= 2:
            return "triclinic"
        elif i <= 15:
            return "monoclinic"
        elif i <= 74:
            return "orthorhombic"
        elif i <= 142:
            return "tetragonal"
        elif i <= 167:
            return "trigonal"
        elif i <= 194:
            return "hexagonal"
        else:
            return "cubic"

    def is_subgroup(self, supergroup):
        """
        True if this space group is a subgroup of the supplied group.

        Args:
            group (Spacegroup): Supergroup to test.

        Returns:
            True if this space group is a subgroup of the supplied group.
        """
        if len(supergroup.symmetry_ops) < len(self.symmetry_ops):
            return False

        groups = [[supergroup.int_number]]
        all_groups = [supergroup.int_number]
        count = 0
        while True:
            new_sub_groups = set()
            for i in groups[-1]:
                new_sub_groups.update([j for j in MAXIMAL_SUBGROUPS[i] if j
                                       not in all_groups])
            if self.int_number in new_sub_groups:
                return True
            elif len(new_sub_groups) == 0:
                break
            else:
                groups.append(new_sub_groups)
                all_groups.extend(new_sub_groups)
        return False

    def is_supergroup(self, subgroup):
        """
        True if this space group is a supergroup of the supplied group.

        Args:
            subgroup (Spacegroup): Subgroup to test.

        Returns:
            True if this space group is a supergroup of the supplied group.
        """
        return subgroup.is_subgroup(self)

    @classmethod
    def from_int_number(cls, int_number, hexagonal=True):
        """
        Obtains a SpaceGroup from its international number.

        Args:
            int_number (int): International number.
            hexagonal (bool): For rhombohedral groups, whether to return the
                hexagonal setting (default) or rhombohedral setting.

        Returns:
            (SpaceGroup)
        """
        return SpaceGroup(sg_symbol_from_int_number(int_number,
                                                    hexagonal=hexagonal))

    def __str__(self):
        return "Spacegroup %s with international number %d and order %d" % (
            self.symbol, self.int_number, len(self.symmetry_ops))


def sg_symbol_from_int_number(int_number, hexagonal=True):
    """
    Obtains a SpaceGroup name from its international number.

    Args:
        int_number (int): International number.
        hexagonal (bool): For rhombohedral groups, whether to return the
            hexagonal setting (default) or rhombohedral setting.

    Returns:
        (str) Spacegroup symbol
    """
    syms = []
    for n, v in SPACE_GROUP_ENC.items():
        if v["int_number"] == int_number:
            syms.append(n)
    if len(syms) == 0:
        raise ValueError("Invalid international number!")
    if len(syms) == 2:
        if hexagonal:
            syms = list(filter(lambda s: s.endswith("H"), syms))
        else:
            syms = list(filter(lambda s: not s.endswith("H"), syms))
    return syms.pop()


def in_array_list(array_list, a, tol=1e-5):
    """
    Extremely efficient nd-array comparison using numpy's broadcasting. This
    function checks if a particular array a, is present in a list of arrays.
    It works for arrays of any size, e.g., even matrix searches.

    Args:
        array_list ([array]): A list of arrays to compare to.
        a (array): The test array for comparison.
        tol (float): The tolerance. Defaults to 1e-5. If 0, an exact match is
            done.

    Returns:
        (bool)
    """
    if len(array_list) == 0:
        return False
    axes = tuple(range(1, a.ndim + 1))
    if not tol:
        return np.any(np.all(np.equal(array_list, a[None, :]), axes))
    else:
        return np.any(np.sum(np.abs(array_list - a[None, :]), axes) < tol)



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

import glob
import os
from io import open
import sys
import platform

from setuptools import setup, find_packages, Extension

try:
    from numpy.distutils.misc_util import get_numpy_include_dirs
except ImportError:
    print("numpy.distutils.misc_util cannot be imported. Please install numpy"
          "first before install pymatgen...")
    sys.exit(-1)

SETUP_PTH = os.path.dirname(__file__)



extra_link_args = []
if sys.platform.startswith('win') and platform.machine().endswith('64'):
    extra_link_args.append('-Wl,--allow-multiple-definition')

with open(os.path.join(SETUP_PTH, "README.rst")) as f:
    long_desc = f.read()
    ind = long_desc.find("\n")
    long_desc = long_desc[ind + 1:]

setup(
    name="pymatgen",
    packages=find_packages(),
    version="4.1.0",
    install_requires=["numpy>=1.9", "six", "atomicfile", "requests",
                      "pybtex", "pyyaml", "monty>=0.7.0", "scipy>=0.14",
                      "tabulate", "enum34", "spglib"],
    extras_require={"plotting": ["matplotlib>=1.1", "prettyplotlib"],
                    "pourbaix diagrams, bandstructure": ["pyhull>=1.5.3"],
                    "ase_adaptor": ["ase>=3.3"],
                    "vis": ["vtk>=6.0.0"],
                    "abinit": ["pydispatcher>=2.0.3", "apscheduler==2.1.0"],
                    "chemenv": ["unittest2"]},
    package_data={"pymatgen.core": ["*.json"],
                  "pymatgen.analysis": ["*.yaml", "*.csv"],
                  "pymatgen.analysis.chemenv.coordination_environments.coordination_geometries_files": ["*.txt",
                                                                                                        "*.json"],
                  "pymatgen.analysis.chemenv.coordination_environments.strategy_files": ["*.json"],
                  "pymatgen.io.vasp": ["*.yaml"],
                  "pymatgen.io.feff": ["*.yaml"],
                  "pymatgen.symmetry": ["*.yaml"],
                  "pymatgen.entries": ["*.yaml"],
                  "pymatgen.structure_prediction": ["data/*.json"],
                  "pymatgen.vis": ["ElementColorSchemes.yaml"],
                  "pymatgen.command_line": ["OxideTersoffPotentials"],
                  "pymatgen.analysis.defects": ["*.json"],
                  "pymatgen.analysis.diffraction": ["*.json"],
                  "pymatgen.util": ["structures/*.json"]},
    author="Pymatgen Development Team",
    author_email="pymatgen@googlegroups.com",
    maintainer="Shyue Ping Ong",
    maintainer_email="ongsp@eng.ucsd.edu",
    url="http://www.pymatgen.org",
    license="MIT",
    description="Python Materials Genomics is a robust materials "
                "analysis code that defines core object representations for "
                "structures and molecules with support for many electronic "
                "structure codes. It is currently the core analysis code "
                "powering the Materials Project "
                "(https://www.materialsproject.org).",
    long_description=long_desc,
    keywords=["VASP", "gaussian", "ABINIT", "nwchem", "materials", "project",
              "electronic", "structure", "analysis", "phase", "diagrams"],
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    ext_modules=[Extension("pymatgen.optimization.linear_assignment",
                           ["pymatgen/optimization/linear_assignment.c"],
                           include_dirs=get_numpy_include_dirs(),
                           extra_link_args=extra_link_args),
                 Extension("pymatgen.util.coord_utils_cython",
                           ["pymatgen/util/coord_utils_cython.c"],
                           include_dirs=get_numpy_include_dirs(),
                           extra_link_args=extra_link_args)],
    scripts=glob.glob(os.path.join(SETUP_PTH, "scripts", "*"))
)



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

import numpy as np

voigt_map = [(0, 0), (1, 1), (2, 2), (1, 2), (2, 0), (0, 1)]
reverse_voigt_map = np.array([[0, 5, 4],
                              [5, 1, 3],
                              [4, 3, 2]])



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import unicode_literals

"""
Interface with command line aconvasp. http://aflowlib.org/
Only tested on Linux. Inspired by Shyue"s qhull_caller
WARNING: you need to have a convasp in your path for this to work
"""

__author__ = "Geoffroy Hautier"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Geoffroy Hautier"
__email__ = "geoffroy.hautier@uclouvain.be"
__status__ = "Production"
__date__ = "$Nov 22, 2011M$"

import subprocess

from pymatgen.io.vasp.inputs import Poscar


def run_aconvasp_command(command, structure):
    """
    Helper function for calling aconvasp with different arguments
    """
    poscar = Poscar(structure)
    p = subprocess.Popen(command, stdout=subprocess.PIPE,
                         stdin=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    output = p.communicate(input=poscar.get_string())
    return output


def get_num_division_kpoints(structure, kppa):
    """
    Get kpoint divisions for a given k-point density (per reciprocal-atom):
    kppa and a given structure
    """
    output = run_aconvasp_command(["aconvasp", "--kpoints", str(kppa)],
                                  structure)
    tmp = output[0].rsplit("\n")[6].rsplit(" ")
    return [int(tmp[5]), int(tmp[6]), int(tmp[7])]


def get_minkowski_red(structure):
    """
    Get a minkowski reduced structure
    """
    output = run_aconvasp_command(["aconvasp", "--kpath"], structure)
    started = False
    poscar_string = ""
    if "ERROR" in output[1]:
        raise AconvaspError(output[1])
    for line in output[0].split("\n"):
        if started or line.find("KPOINTS TO RUN") != -1:
            poscar_string = poscar_string + line + "\n"
        if line.find("STRUCTURE TO RUN") != -1:
            started = True
        if line.find("KPOINTS TO RUN") != -1:
            started = False
    return Poscar.from_string(poscar_string).structure


def get_conv_struct(structure):
    """
    Get a minkowski reduced structure
    """
    output = run_aconvasp_command(["aconvasp", "--std_conv"], structure)
    if "ERROR" in output[1]:
        raise AconvaspError(output[1])
    tmp = Poscar.from_string(output[0])
    return {'struct': tmp.structure, 'comm': tmp.comment}


def get_prim_struct(structure):
    """
    Get standard primitive
    """
    output = run_aconvasp_command(["aconvasp", "--std_prim"], structure)
    if "ERROR" in output[1]:
        raise AconvaspError(output[1])
    tmp = Poscar.from_string(output[0])
    return {'struct': tmp.structure, 'comm': tmp.comment}


def get_vasp_kpoint_file_sym(structure):
    """
    get a kpoint file ready to be ran in VASP along the symmetry lines of the
    Brillouin Zone
    """
    output = run_aconvasp_command(["aconvasp", "--kpath"], structure)
    if "ERROR" in output[1]:
        raise AconvaspError(output[1])
    started = False
    kpoints_string = ""
    for line in output[0].split("\n"):
        #print line
        if started or line.find("END") != -1:
            kpoints_string = kpoints_string + line + "\n"
        if line.find("KPOINTS TO RUN") != -1:
            started = True
        if line.find("END") != -1:
            started = False
    return kpoints_string


class AconvaspError(Exception):
    """
    Exception class for aconvasp.
    Raised when the aconvasp gives an error
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return "AconvaspError : " + self.msg



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals

"""
This module contains classes to wrap Python VTK to make nice molecular plots.
"""


__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__date__ = "Nov 27, 2011"


import os
import itertools
import math
import subprocess

import numpy as np
try:
    import vtk
    from vtk import vtkInteractorStyleTrackballCamera
except ImportError:
    # VTK not present. The Camera is to set object to avoid errors in unittest.
    vtk = None
    vtkInteractorStyleTrackballCamera = object

from monty.serialization import loadfn
from monty.dev import requires

from pymatgen.util.coord_utils import in_coord_list
from pymatgen.core.periodic_table import Specie
from pymatgen.core.structure import Structure


module_dir = os.path.dirname(os.path.abspath(__file__))
EL_COLORS = loadfn(os.path.join(module_dir, "ElementColorSchemes.yaml"))


class StructureVis(object):
    """
    Provides Structure object visualization using VTK.
    """

    @requires(vtk, "Visualization requires the installation of VTK with "
                   "Python bindings.")
    def __init__(self, element_color_mapping=None, show_unit_cell=True,
                 show_bonds=False, show_polyhedron=True,
                 poly_radii_tol_factor=0.5, excluded_bonding_elements=None):
        """
        Constructs a Structure Visualization.

        Args:
            element_color_mapping: Optional color mapping for the elements,
                as a dict of {symbol: rgb tuple}. For example, {"Fe": (255,
                123,0), ....} If None is specified, a default based on
                Jmol"s color scheme is used.
            show_unit_cell: Set to False to not show the unit cell
                boundaries. Defaults to True.
            show_bonds: Set to True to show bonds. Defaults to True.
            show_polyhedron: Set to True to show polyhedrons. Defaults to
                False.
            poly_radii_tol_factor: The polyhedron and bonding code uses the
                ionic radii of the elements or species to determine if two
                atoms are bonded. This specifies a tolerance scaling factor
                such that atoms which are (1 + poly_radii_tol_factor) * sum
                of ionic radii apart are still considered as bonded.
            excluded_bonding_elements: List of atom types to exclude from
                bonding determination. Defaults to an empty list. Useful
                when trying to visualize a certain atom type in the
                framework (e.g., Li in a Li-ion battery cathode material).

        Useful keyboard shortcuts implemented.
            h : Show help
            A/a : Increase/decrease cell by one unit vector in a-direction
            B/b : Increase/decrease cell by one unit vector in b-direction
            C/c : Increase/decrease cell by one unit vector in c-direction
            # : Toggle showing of polyhedrons
            - : Toggle showing of bonds
            [ : Decrease poly_radii_tol_factor by 0.05
            ] : Increase poly_radii_tol_factor by 0.05
            r : Reset camera direction
            o : Orthogonalize structure
            Up/Down : Rotate view along Up direction by 90 clock/anticlockwise
            Left/right : Rotate view along camera direction by 90
            clock/anticlockwise
        """
        # create a rendering window and renderer
        self.ren = vtk.vtkRenderer()
        self.ren_win = vtk.vtkRenderWindow()
        self.ren_win.AddRenderer(self.ren)
        self.ren.SetBackground(1, 1, 1)
        self.title = "Structure Visualizer"
        # create a renderwindowinteractor
        self.iren = vtk.vtkRenderWindowInteractor()
        self.iren.SetRenderWindow(self.ren_win)
        self.mapper_map = {}
        self.structure = None

        if element_color_mapping:
            self.el_color_mapping = element_color_mapping
        else:
            self.el_color_mapping = EL_COLORS["VESTA"]
        self.show_unit_cell = show_unit_cell
        self.show_bonds = show_bonds
        self.show_polyhedron = show_polyhedron
        self.poly_radii_tol_factor = poly_radii_tol_factor
        self.excluded_bonding_elements = excluded_bonding_elements if \
            excluded_bonding_elements else []
        self.show_help = True
        self.supercell = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        self.redraw()

        style = StructureInteractorStyle(self)
        self.iren.SetInteractorStyle(style)
        self.ren.parent = self

    def rotate_view(self, axis_ind=0, angle=0):
        """
        Rotate the camera view.

        Args:
            axis_ind: Index of axis to rotate. Defaults to 0, i.e., a-axis.
            angle: Angle to rotate by. Defaults to 0.
        """
        camera = self.ren.GetActiveCamera()
        if axis_ind == 0:
            camera.Roll(angle)
        elif axis_ind == 1:
            camera.Azimuth(angle)
        else:
            camera.Pitch(angle)
        self.ren_win.Render()

    def write_image(self, filename="image.png", magnification=1,
                    image_format="png"):
        """
        Save render window to an image.

        Arguments:
            filename:
                filename to save to. Defaults to image.png.
            magnification:
                magnification. Use it to render high res images.
            image_format:
                choose between jpeg, png.  Png is the default.
        """
        render_large = vtk.vtkRenderLargeImage()
        render_large.SetInput(self.ren)
        if image_format == "jpeg":
            writer = vtk.vtkJPEGWriter()
            writer.SetQuality(80)
        else:
            writer = vtk.vtkPNGWriter()

        render_large.SetMagnification(magnification)
        writer.SetFileName(filename)

        writer.SetInputConnection(render_large.GetOutputPort())
        self.ren_win.Render()
        writer.Write()
        del render_large

    def redraw(self, reset_camera=False):
        """
        Redraw the render window.

        Args:
            reset_camera: Set to True to reset the camera to a
                pre-determined default for each structure.  Defaults to False.
        """
        self.ren.RemoveAllViewProps()
        self.picker = None
        self.add_picker_fixed()
        self.helptxt_mapper = vtk.vtkTextMapper()
        tprops = self.helptxt_mapper.GetTextProperty()
        tprops.SetFontSize(14)
        tprops.SetFontFamilyToTimes()
        tprops.SetColor(0, 0, 0)

        if self.structure is not None:
            self.set_structure(self.structure, reset_camera)

        self.ren_win.Render()

    def orthongonalize_structure(self):
        if self.structure is not None:
            self.set_structure(self.structure.copy(sanitize=True))
        self.ren_win.Render()

    def display_help(self):
        """
        Display the help for various keyboard shortcuts.
        """
        helptxt = ["h : Toggle help",
                   "A/a, B/b or C/c : Increase/decrease cell by one a,"
                   " b or c unit vector", "# : Toggle showing of polyhedrons",
                   "-: Toggle showing of bonds", "r : Reset camera direction",
                   "[/]: Decrease or increase poly_radii_tol_factor "
                   "by 0.05. Value = " + str(self.poly_radii_tol_factor),
                   "Up/Down: Rotate view along Up direction by 90 "
                   "clockwise/anticlockwise",
                   "Left/right: Rotate view along camera direction by "
                   "90 clockwise/anticlockwise", "s: Save view to image.png",
                   "o: Orthogonalize structure"]
        self.helptxt_mapper.SetInput("\n".join(helptxt))
        self.helptxt_actor.SetPosition(10, 10)
        self.helptxt_actor.VisibilityOn()

    def set_structure(self, structure, reset_camera=True):
        """
        Add a structure to the visualizer.

        Args:
            structure: structure to visualize
            reset_camera: Set to True to reset the camera to a default
                determined based on the structure.
        """
        self.ren.RemoveAllViewProps()

        has_lattice = hasattr(structure, "lattice")

        if has_lattice:
            s = Structure.from_sites(structure, to_unit_cell=True)
            s.make_supercell(self.supercell)
        else:
            s = structure

        inc_coords = []
        for site in s:
            self.add_site(site)
            inc_coords.append(site.coords)

        count = 0
        labels = ["a", "b", "c"]
        colors = [(1, 0, 0), (0, 1, 0), (0, 0, 1)]

        if has_lattice:
            matrix = s.lattice.matrix

        if self.show_unit_cell and has_lattice:
            #matrix = s.lattice.matrix
            self.add_text([0, 0, 0], "o")
            for vec in matrix:
                self.add_line((0, 0, 0), vec, colors[count])
                self.add_text(vec, labels[count], colors[count])
                count += 1
            for (vec1, vec2) in itertools.permutations(matrix, 2):
                self.add_line(vec1, vec1 + vec2)
            for (vec1, vec2, vec3) in itertools.permutations(matrix, 3):
                self.add_line(vec1 + vec2, vec1 + vec2 + vec3)

        if self.show_bonds or self.show_polyhedron:
            elements = sorted(s.composition.elements, key=lambda a: a.X)
            anion = elements[-1]

            def contains_anion(site):
                for sp in site.species_and_occu.keys():
                    if sp.symbol == anion.symbol:
                        return True
                return False

            anion_radius = anion.average_ionic_radius
            for site in s:
                exclude = False
                max_radius = 0
                color = np.array([0, 0, 0])
                for sp, occu in site.species_and_occu.items():
                    if sp.symbol in self.excluded_bonding_elements \
                            or sp == anion:
                        exclude = True
                        break
                    max_radius = max(max_radius, sp.average_ionic_radius)
                    color = color + \
                        occu * np.array(self.el_color_mapping.get(sp.symbol,
                                                                  [0, 0, 0]))

                if not exclude:
                    max_radius = (1 + self.poly_radii_tol_factor) * \
                        (max_radius + anion_radius)
                    nn = structure.get_neighbors(site, float(max_radius))
                    nn_sites = []
                    for nnsite, dist in nn:
                        if contains_anion(nnsite):
                            nn_sites.append(nnsite)
                            if not in_coord_list(inc_coords, nnsite.coords):
                                self.add_site(nnsite)
                    if self.show_bonds:
                        self.add_bonds(nn_sites, site)
                    if self.show_polyhedron:
                        color = [i / 255 for i in color]
                        self.add_polyhedron(nn_sites, site, color)

        if self.show_help:
            self.helptxt_actor = vtk.vtkActor2D()
            self.helptxt_actor.VisibilityOn()
            self.helptxt_actor.SetMapper(self.helptxt_mapper)
            self.ren.AddActor(self.helptxt_actor)
            self.display_help()

        camera = self.ren.GetActiveCamera()
        if reset_camera:
            if has_lattice:
                #Adjust the camera for best viewing
                lengths = s.lattice.abc
                pos = (matrix[1] + matrix[2]) * 0.5 + \
                    matrix[0] * max(lengths) / lengths[0] * 3.5
                camera.SetPosition(pos)
                camera.SetViewUp(matrix[2])
                camera.SetFocalPoint((matrix[0] + matrix[1] + matrix[2]) * 0.5)
            else:
                origin = s.center_of_mass
                max_site = max(
                    s, key=lambda site: site.distance_from_point(origin))
                camera.SetPosition(origin + 5 * (max_site.coords - origin))
                camera.SetFocalPoint(s.center_of_mass)

        self.structure = structure
        self.title = s.composition.formula

    def zoom(self, factor):
        """
        Zoom the camera view by a factor.
        """
        camera = self.ren.GetActiveCamera()
        camera.Zoom(factor)
        self.ren_win.Render()

    def show(self):
        """
        Display the visualizer.
        """
        self.iren.Initialize()
        self.ren_win.SetSize(800, 800)
        self.ren_win.SetWindowName(self.title)
        self.ren_win.Render()
        self.iren.Start()

    def add_site(self, site):
        """
        Add a site to the render window. The site is displayed as a sphere, the
        color of which is determined based on the element. Partially occupied
        sites are displayed as a single element color, though the site info
        still shows the partial occupancy.

        Args:
            site: Site to add.
        """
        start_angle = 0
        radius = 0
        total_occu = 0

        for specie, occu in site.species_and_occu.items():
            radius += occu * (specie.ionic_radius
                              if isinstance(specie, Specie)
                              and specie.ionic_radius
                              else specie.average_ionic_radius)
            total_occu += occu

        for specie, occu in site.species_and_occu.items():
            if not specie:
                color = (1, 1, 1)
            elif specie.symbol in self.el_color_mapping:
                color = [i / 255 for i in self.el_color_mapping[specie.symbol]]
            mapper = self.add_partial_sphere(site.coords, radius, color,
                start_angle, start_angle + 360 * occu)
            self.mapper_map[mapper] = [site]
            start_angle += 360 * occu

        if total_occu < 1:
            mapper = self.add_partial_sphere(site.coords, radius, (1,1,1),
                start_angle, start_angle + 360 * (1 - total_occu))
            self.mapper_map[mapper] = [site]

    def add_partial_sphere(self, coords, radius, color, start=0, end=360,
                           opacity=1):
        sphere = vtk.vtkSphereSource()
        sphere.SetCenter(coords)
        sphere.SetRadius(0.2 + 0.002 * radius)
        sphere.SetThetaResolution(18)
        sphere.SetPhiResolution(18)
        sphere.SetStartTheta(start)
        sphere.SetEndTheta(end)

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(sphere.GetOutputPort())
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(color)
        actor.GetProperty().SetOpacity(opacity)
        self.ren.AddActor(actor)
        return mapper

    def add_text(self, coords, text, color=(0, 0, 0)):
        """
        Add text at a coordinate.

        Args:
            coords: Coordinates to add text at.
            text: Text to place.
            color: Color for text as RGB. Defaults to black.
        """
        source = vtk.vtkVectorText()
        source.SetText(text)
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(source.GetOutputPort())
        follower = vtk.vtkFollower()
        follower.SetMapper(mapper)
        follower.GetProperty().SetColor(color)
        follower.SetPosition(coords)
        follower.SetScale(0.5)
        self.ren.AddActor(follower)
        follower.SetCamera(self.ren.GetActiveCamera())

    def add_line(self, start, end, color=(0.5, 0.5, 0.5), width=1):
        """
        Adds a line.

        Args:
            start: Starting coordinates for line.
            end: Ending coordinates for line.
            color: Color for text as RGB. Defaults to grey.
            width: Width of line. Defaults to 1.
        """
        source = vtk.vtkLineSource()
        source.SetPoint1(start)
        source.SetPoint2(end)

        vertexIDs = vtk.vtkStringArray()
        vertexIDs.SetNumberOfComponents(1)
        vertexIDs.SetName("VertexIDs")
        # Set the vertex labels
        vertexIDs.InsertNextValue("a")
        vertexIDs.InsertNextValue("b")
        source.GetOutput().GetPointData().AddArray(vertexIDs)

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(source.GetOutputPort())
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(color)
        actor.GetProperty().SetLineWidth(width)
        self.ren.AddActor(actor)

    def add_polyhedron(self, neighbors, center, color, opacity=1.0,
                       draw_edges=False, edges_color=[0.0, 0.0, 0.0],
                       edges_linewidth=2):
        """
        Adds a polyhedron.

        Args:
            neighbors: Neighbors of the polyhedron (the vertices).
            center: The atom in the center of the polyhedron.
            color: Color for text as RGB.
            opacity: Opacity of the polyhedron
            draw_edges: If set to True, the a line will be drawn at each edge
            edges_color: Color of the line for the edges
            edges_linewidth: Width of the line drawn for the edges
        """
        points = vtk.vtkPoints()
        conv = vtk.vtkConvexPointSet()
        for i in range(len(neighbors)):
            x, y, z = neighbors[i].coords
            points.InsertPoint(i, x, y, z)
            conv.GetPointIds().InsertId(i, i)
        grid = vtk.vtkUnstructuredGrid()
        grid.Allocate(1, 1)
        grid.InsertNextCell(conv.GetCellType(), conv.GetPointIds())
        grid.SetPoints(points)

        dsm = vtk.vtkDataSetMapper()
        polysites = [center]
        polysites.extend(neighbors)
        self.mapper_map[dsm] = polysites
        if vtk.VTK_MAJOR_VERSION <= 5:
            dsm.SetInputConnection(grid.GetProducerPort())
        else:
            dsm.SetInputData(grid)
        ac = vtk.vtkActor()
        #ac.SetMapper(mapHull)
        ac.SetMapper(dsm)
        ac.GetProperty().SetOpacity(opacity)
        if color == 'element':
            # If partial occupations are involved, the color of the specie with
            # the highest occupation is used
            myoccu = 0.0
            for specie, occu in center.species_and_occu.items():
                if occu > myoccu:
                    myspecie = specie
                    myoccu = occu
            color = [i / 255 for i in self.el_color_mapping[myspecie.symbol]]
            ac.GetProperty().SetColor(color)
        else:
            ac.GetProperty().SetColor(color)
        if draw_edges:
            ac.GetProperty().SetEdgeColor(edges_color)
            ac.GetProperty().SetLineWidth(edges_linewidth)
            ac.GetProperty().EdgeVisibilityOn()
        self.ren.AddActor(ac)

    def add_triangle(self, neighbors, color, center=None, opacity=0.4,
                     draw_edges=False, edges_color=[0.0, 0.0, 0.0],
                     edges_linewidth=2):
        """
        Adds a triangular surface between three atoms.

        Args:
            atoms: Atoms between which a triangle will be drawn.
            color: Color for triangle as RGB.
            center: The "central atom" of the triangle
            opacity: opacity of the triangle
            draw_edges: If set to True, the a line will be  drawn at each edge
            edges_color: Color of the line for the edges
            edges_linewidth: Width of the line drawn for the edges
        """
        points = vtk.vtkPoints()
        triangle = vtk.vtkTriangle()
        for ii in range(3):
            points.InsertNextPoint(neighbors[ii].x, neighbors[ii].y,
                                   neighbors[ii].z)
            triangle.GetPointIds().SetId(ii, ii)
        triangles = vtk.vtkCellArray()
        triangles.InsertNextCell(triangle)

        # polydata object
        trianglePolyData = vtk.vtkPolyData()
        trianglePolyData.SetPoints( points )
        trianglePolyData.SetPolys( triangles )

        # mapper
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInput(trianglePolyData)

        ac = vtk.vtkActor()
        ac.SetMapper(mapper)
        ac.GetProperty().SetOpacity(opacity)
        if color == 'element':
            if center is None:
                raise ValueError(
                    'Color should be chosen according to the central atom, '
                    'and central atom is not provided')
            # If partial occupations are involved, the color of the specie with
            # the highest occupation is used
            myoccu = 0.0
            for specie, occu in center.species_and_occu.items():
                if occu > myoccu:
                    myspecie = specie
                    myoccu = occu
            color = [i / 255 for i in self.el_color_mapping[myspecie.symbol]]
            ac.GetProperty().SetColor(color)
        else:
            ac.GetProperty().SetColor(color)
        if draw_edges:
            ac.GetProperty().SetEdgeColor(edges_color)
            ac.GetProperty().SetLineWidth(edges_linewidth)
            ac.GetProperty().EdgeVisibilityOn()
        self.ren.AddActor(ac)

    def add_faces(self, faces, color, opacity=0.35):
        for face in faces:
            if len(face) == 3:
                points = vtk.vtkPoints()
                triangle = vtk.vtkTriangle()
                for ii in range(3):
                    points.InsertNextPoint(face[ii][0], face[ii][1], face[ii][2])
                    triangle.GetPointIds().SetId(ii, ii)
                triangles = vtk.vtkCellArray()
                triangles.InsertNextCell(triangle)
                trianglePolyData = vtk.vtkPolyData()
                trianglePolyData.SetPoints(points)
                trianglePolyData.SetPolys(triangles)
                mapper = vtk.vtkPolyDataMapper()
                if vtk.VTK_MAJOR_VERSION <= 5:
                    mapper.SetInputConnection(trianglePolyData.GetProducerPort())
                else:
                    mapper.SetInputData(trianglePolyData)
                # mapper.SetInput(trianglePolyData)
                ac = vtk.vtkActor()
                ac.SetMapper(mapper)
                ac.GetProperty().SetOpacity(opacity)
                ac.GetProperty().SetColor(color)
                self.ren.AddActor(ac)
            elif False and len(face) == 4:
                points = vtk.vtkPoints()
                for ii in range(4):
                    points.InsertNextPoint(face[ii][0], face[ii][1], face[ii][2])
                line1 = vtk.vtkLine()
                line1.GetPointIds().SetId(0, 0)
                line1.GetPointIds().SetId(1, 2)
                line2 = vtk.vtkLine()
                line2.GetPointIds().SetId(0, 3)
                line2.GetPointIds().SetId(1, 1)
                lines = vtk.vtkCellArray()
                lines.InsertNextCell(line1)
                lines.InsertNextCell(line2)
                polydata = vtk.vtkPolyData()
                polydata.SetPoints(points)
                polydata.SetLines(lines)
                ruledSurfaceFilter = vtk.vtkRuledSurfaceFilter()
                ruledSurfaceFilter.SetInput(polydata)
                ruledSurfaceFilter.SetResolution(15, 15)
                ruledSurfaceFilter.SetRuledModeToResample()
                mapper = vtk.vtkPolyDataMapper()
                mapper.SetInput(ruledSurfaceFilter.GetOutput())
                ac = vtk.vtkActor()
                ac.SetMapper(mapper)
                ac.GetProperty().SetOpacity(opacity)
                ac.GetProperty().SetColor(color)
                self.ren.AddActor(ac)
            elif len(face) > 3:
                center = np.zeros(3, np.float)
                for site in face:
                    center += site
                center /= np.float(len(face))
                for ii in range(len(face)):
                    points = vtk.vtkPoints()
                    triangle = vtk.vtkTriangle()
                    points.InsertNextPoint(face[ii][0], face[ii][1], face[ii][2])
                    ii2 = np.mod(ii+1, len(face))
                    points.InsertNextPoint(face[ii2][0], face[ii2][1], face[ii2][2])
                    points.InsertNextPoint(center[0], center[1], center[2])
                    for ii in range(3):
                        triangle.GetPointIds().SetId(ii, ii)
                    triangles = vtk.vtkCellArray()
                    triangles.InsertNextCell(triangle)
                    trianglePolyData = vtk.vtkPolyData()
                    trianglePolyData.SetPoints(points)
                    trianglePolyData.SetPolys(triangles)
                    mapper = vtk.vtkPolyDataMapper()
                    if vtk.VTK_MAJOR_VERSION <= 5:
                        mapper.SetInputConnection(trianglePolyData.GetProducerPort())
                    else:
                        mapper.SetInputData(trianglePolyData)
                    # mapper.SetInput(trianglePolyData)
                    ac = vtk.vtkActor()
                    ac.SetMapper(mapper)
                    ac.GetProperty().SetOpacity(opacity)
                    ac.GetProperty().SetColor(color)
                    self.ren.AddActor(ac)
            else:
                raise ValueError("Number of points for a face should be >= 3")

    def add_edges(self, edges, type='line', linewidth=2, color=[0.0, 0.0, 0.0]):
        points = vtk.vtkPoints()
        lines = vtk.vtkCellArray()
        for iedge, edge in enumerate(edges):
            points.InsertPoint(2*iedge, edge[0])
            points.InsertPoint(2*iedge + 1, edge[1])
            lines.InsertNextCell(2)
            lines.InsertCellPoint(2*iedge)
            lines.InsertCellPoint(2*iedge + 1)
        polydata = vtk.vtkPolyData()
        polydata.SetPoints(points)
        polydata.SetLines(lines)
        mapper = vtk.vtkPolyDataMapper()
        if vtk.VTK_MAJOR_VERSION <= 5:
            mapper.SetInputConnection(polydata.GetProducerPort())
        else:
            mapper.SetInputData(polydata)
        # mapper.SetInput(polydata)
        ac = vtk.vtkActor()
        ac.SetMapper(mapper)
        ac.GetProperty().SetColor(color)
        ac.GetProperty().SetLineWidth(linewidth)
        self.ren.AddActor(ac)

    def add_bonds(self, neighbors, center, color=None, opacity=None,
                  radius=0.1):
        """
        Adds bonds for a site.

        Args:
            neighbors: Neighbors of the site.
            center: The site in the center for all bonds.
            color: Color of the tubes representing the bonds
            opacity: Opacity of the tubes representing the bonds
            radius: Radius of tube s representing the bonds
        """
        points = vtk.vtkPoints()
        points.InsertPoint(0, center.x, center.y, center.z)
        n = len(neighbors)
        lines = vtk.vtkCellArray()
        for i in range(n):
            points.InsertPoint(i + 1, neighbors[i].coords)
            lines.InsertNextCell(2)
            lines.InsertCellPoint(0)
            lines.InsertCellPoint(i + 1)
        pd = vtk.vtkPolyData()
        pd.SetPoints(points)
        pd.SetLines(lines)

        tube = vtk.vtkTubeFilter()
        if vtk.VTK_MAJOR_VERSION <= 5:
            tube.SetInputConnection(pd.GetProducerPort())
        else:
            tube.SetInputData(pd)
        tube.SetRadius(radius)

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(tube.GetOutputPort())

        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        if opacity is not None:
            actor.GetProperty().SetOpacity(opacity)
        if color is not None:
            actor.GetProperty().SetColor(color)
        self.ren.AddActor(actor)

    def add_picker_fixed(self):
        # Create a cell picker.
        picker = vtk.vtkCellPicker()
        # Create a Python function to create the text for the text mapper used
        # to display the results of picking.

        def annotate_pick(obj, event):
            if picker.GetCellId() < 0 and not self.show_help:
                self.helptxt_actor.VisibilityOff()
            else:
                mapper = picker.GetMapper()
                if mapper in self.mapper_map:
                    output = []
                    for site in self.mapper_map[mapper]:
                        row = ["{} - ".format(site.species_string),
                               ", ".join(["{:.3f}".format(c)
                                          for c in site.frac_coords]),
                               "[" + ", ".join(["{:.3f}".format(c)
                                                for c in site.coords]) +
                               "]"]
                        output.append("".join(row))
                    self.helptxt_mapper.SetInput("\n".join(output))
                    self.helptxt_actor.SetPosition(10, 10)
                    self.helptxt_actor.VisibilityOn()
                    self.show_help = False
        self.picker = picker
        picker.AddObserver("EndPickEvent", annotate_pick)
        self.iren.SetPicker(picker)

    def add_picker(self):
        # Create a cell picker.
        picker = vtk.vtkCellPicker()
        # Create a Python function to create the text for the text mapper used
        # to display the results of picking.
        source = vtk.vtkVectorText()
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(source.GetOutputPort())
        follower = vtk.vtkFollower()
        follower.SetMapper(mapper)
        follower.GetProperty().SetColor((0, 0, 0))
        follower.SetScale(0.2)
        self.ren.AddActor(follower)
        follower.SetCamera(self.ren.GetActiveCamera())
        follower.VisibilityOff()

        def annotate_pick(obj, event):
            if picker.GetCellId() < 0:
                follower.VisibilityOff()
            else:
                pick_pos = picker.GetPickPosition()
                mapper = picker.GetMapper()
                if mapper in self.mapper_map:
                    site = self.mapper_map[mapper]
                    output = [site.species_string, "Frac. coords: " +
                                                   " ".join(["{:.4f}".format(c)
                                                             for c in
                                                             site.frac_coords])]
                    source.SetText("\n".join(output))
                    follower.SetPosition(pick_pos)
                    follower.VisibilityOn()
        picker.AddObserver("EndPickEvent", annotate_pick)
        self.picker = picker
        self.iren.SetPicker(picker)


class StructureInteractorStyle(vtkInteractorStyleTrackballCamera):
    """
    A custom interactor style for visualizing structures.
    """

    def __init__(self, parent):
        self.parent = parent
        self.AddObserver("LeftButtonPressEvent", self.leftButtonPressEvent)
        self.AddObserver("MouseMoveEvent", self.mouseMoveEvent)
        self.AddObserver("LeftButtonReleaseEvent", self.leftButtonReleaseEvent)
        self.AddObserver("KeyPressEvent", self.keyPressEvent)

    def leftButtonPressEvent(self, obj, event):
        self.mouse_motion = 0
        self.OnLeftButtonDown()
        return

    def mouseMoveEvent(self, obj, event):
        self.mouse_motion = 1
        self.OnMouseMove()
        return

    def leftButtonReleaseEvent(self, obj, event):
        ren = obj.GetCurrentRenderer()
        iren = ren.GetRenderWindow().GetInteractor()
        if self.mouse_motion == 0:
            pos = iren.GetEventPosition()
            iren.GetPicker().Pick(pos[0], pos[1], 0, ren)
        self.OnLeftButtonUp()
        return

    def keyPressEvent(self, obj, event):
        parent = obj.GetCurrentRenderer().parent
        sym = parent.iren.GetKeySym()

        if sym in "ABCabc":
            if sym == "A":
                parent.supercell[0][0] += 1
            elif sym == "B":
                parent.supercell[1][1] += 1
            elif sym == "C":
                parent.supercell[2][2] += 1
            elif sym == "a":
                parent.supercell[0][0] = max(parent.supercell[0][0] - 1, 1)
            elif sym == "b":
                parent.supercell[1][1] = max(parent.supercell[1][1] - 1, 1)
            elif sym == "c":
                parent.supercell[2][2] = max(parent.supercell[2][2] - 1, 1)
            parent.redraw()
        elif sym == "numbersign":
            parent.show_polyhedron = not parent.show_polyhedron
            parent.redraw()
        elif sym == "minus":
            parent.show_bonds = not parent.show_bonds
            parent.redraw()
        elif sym == "bracketleft":
            parent.poly_radii_tol_factor -= 0.05 \
                if parent.poly_radii_tol_factor > 0 else 0
            parent.redraw()
        elif sym == "bracketright":
            parent.poly_radii_tol_factor += 0.05
            parent.redraw()
        elif sym == "h":
            parent.show_help = not parent.show_help
            parent.redraw()
        elif sym == "r":
            parent.redraw(True)
        elif sym == "s":
            parent.write_image("image.png")
        elif sym == "Up":
            parent.rotate_view(1, 90)
        elif sym == "Down":
            parent.rotate_view(1, -90)
        elif sym == "Left":
            parent.rotate_view(0, -90)
        elif sym == "Right":
            parent.rotate_view(0, 90)
        elif sym == "o":
            parent.orthongonalize_structure()
            parent.redraw()

        self.OnKeyPress()


def make_movie(structures, output_filename="movie.mp4", zoom=1.0, fps=20,
               bitrate="10000k", quality=1, **kwargs):
    """
    Generate a movie from a sequence of structures using vtk and ffmpeg.

    Args:
        structures ([Structure]): sequence of structures
        output_filename (str): filename for structure output. defaults to
            movie.mp4
        zoom (float): A zoom to be applied to the visualizer. Defaults to 1.0.
        fps (int): Frames per second for the movie. Defaults to 20.
        bitrate (str): Video bitate.  Defaults to "10000k" (fairly high
            quality).
        quality (int): A quality scale. Defaults to 1.
        \*\*kwargs: Any kwargs supported by StructureVis to modify the images
            generated.
    """
    vis = StructureVis(**kwargs)
    vis.show_help = False
    vis.redraw()
    vis.zoom(zoom)
    sigfig = int(math.floor(math.log10(len(structures))) + 1)
    filename = "image{0:0" + str(sigfig) + "d}.png"
    for i, s in enumerate(structures):
        vis.set_structure(s)
        vis.write_image(filename.format(i), 3)
    filename = "image%0" + str(sigfig) + "d.png"
    args = ["ffmpeg", "-y", "-i", filename,
            "-q:v", str(quality), "-r", str(fps), "-b:v", str(bitrate),
            output_filename]
    subprocess.Popen(args)
